from jolt import attributes as jolt_attributes
from jolt import filesystem as fs
from jolt import utils
from jolt import Task


from jolt.pkgs import golang


class attributes(object):
    @staticmethod
    def flags(attrib):
        """
        Decorates a task with an alternative ``flags`` attribute.

        The new attribute will be concatenated with the regular
        ``flags`` attribute.

        Args:
            attrib (str): Name of alternative attribute.
                Keywords are expanded.
        """
        return utils.concat_attributes("flags", attrib)


@attributes.flags("flags")
@jolt_attributes.requires("requires_go")
class Executable(Task):
    """ Builds a Go executable """

    abstract = True

    binary = "{canonical_name}"
    """ Name of the target binary (defaults to canonical task name) """

    flags = []
    """ List of build flags, passed directly to 'go build' """

    publishdir = "bin/"
    """ The artifact path where the executable is published. """

    requires_go = ["golang"]
    """
    Go toolchain requirement.

    Override to select another version or implementation.

    Example:

      .. code-block:: python

        requires_go = ["golang:version=1.16.4"]

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.binary = self.expand(self.binary)
        self.publishdir = self.expand(self.publishdir)
        golang  # Dummy reference

    def clean(self, tools):
        self.cachedir = tools.builddir("cache", incremental=True)
        tools.rmtree(self.cachedir, onerror=fs.onerror_warning)

    def run(self, deps, tools):
        self.cachedir = tools.builddir("cache", incremental=True)
        self.outdir = tools.builddir()
        self.info("Building {binary}")
        with tools.environ(GOCACHE=self.cachedir):
            tools.run("go build -o {outdir}/{binary} {}", " ".join(self._flags()))

    def publish(self, artifact, tools):
        with tools.cwd(self.outdir):
            artifact.collect("*", self.publishdir)



from collections import OrderedDict
from contextlib import contextmanager
import fasteners
import json
import glob
from importlib.machinery import SourceFileLoader
import os
import platform
import subprocess
import sys
from types import ModuleType

from jolt import inspection
from jolt.tasks import attributes
from jolt.tasks import Alias, Task, TaskGenerator, TaskRegistry, WorkspaceResource
from jolt.error import raise_error_if, raise_task_error_if
from jolt import common_pb2 as common_pb
from jolt import config
from jolt import filesystem as fs
from jolt import log
from jolt import utils
from jolt.manifest import ManifestExtension
from jolt.manifest import ManifestExtensionRegistry
from jolt.tools import Tools


class Recipe(object):
    def __init__(self, path, joltdir=None, project=None, source=None):
        self.path = path
        self.basepath = os.path.basename(path)
        self.joltdir = joltdir
        self.project = project
        self.source = source
        self.tasks = []

    def load(self, joltdir=None):
        raise_error_if(self.source is not None, "recipe already loaded: {}", self.path)

        with open(self.path) as f:
            self.source = f.read()

    def save(self):
        raise_error_if(self.source is None, "recipe source unknown: {}", self.path)

        with open(self.path, "w") as f:
            f.write(self.source)


class NativeRecipe(Recipe):
    @staticmethod
    def _is_abstract(cls):
        return cls.__dict__.get("abstract", False) or cls.__name__.startswith("_")

    @staticmethod
    def _is_task(cls):
        return isinstance(cls, type) and \
            issubclass(cls, Task) and \
            not NativeRecipe._is_abstract(cls)

    def load(self, joltdir=None):
        super(NativeRecipe, self).load()

        name = utils.canonical(self.path)
        loader = SourceFileLoader("joltfile_{0}".format(name), self.path)
        module = ModuleType(loader.name)
        module.__file__ = self.path
        loader.exec_module(module)
        sys.modules[loader.name] = module

        classes = inspection.getmoduleclasses(module, [Task, TaskGenerator], NativeRecipe._is_abstract)
        generators = []

        for cls in classes[TaskGenerator]:
            cls.joltdir = os.path.normpath(joltdir or self.joltdir or os.path.dirname(self.path))
            generators.append(cls())

        for generator in generators:
            generated_tasks = utils.as_list(generator.generate())
            classes[Task] += filter(NativeRecipe._is_task, generated_tasks)

        for task in classes[Task]:
            task.name = task.name or task.__name__.lower()
            task.joltdir = os.path.normpath(joltdir or self.joltdir or os.path.dirname(self.path))
            task.joltproject = self.project
            self.tasks.append(task)

        log.verbose("Loaded: {0}", self.path)


class Loader(object):
    def recipes(self):
        pass


class LoaderFactory(object):
    def create(self):
        raise NotImplementedError()


class NativeLoader(Loader):
    def __init__(self, searchpath):
        self._recipes = []
        self._find_files(searchpath)

    def _find_files(self, searchpath):
        # If the searchpath is a file, load it directly
        if fs.path.isdir(searchpath):
            return self._load_files(glob.glob(fs.path.join(searchpath, "*.jolt")))

        _, ext = fs.path.splitext(searchpath)
        if ext in [".jolt", ".py"]:
            return self._load_files([searchpath])

    def _load_files(self, files):
        for filepath in files:
            recipe = NativeRecipe(filepath)
            self._recipes.append(recipe)

    @property
    def recipes(self):
        return self._recipes


_loaders = []


def register(factory):
    raise_error_if(not issubclass(factory, LoaderFactory),
                   "{} is not a LoaderFactory", factory.__name__)
    _loaders.append(factory)


@register
class NativeLoaderFactory(LoaderFactory):
    def create(self, searchpath):
        return NativeLoader(searchpath)


@utils.Singleton
class JoltLoader(object):
    filename = "*.jolt"

    def __init__(self):
        self._lock = None
        self._recipes = []
        self._tasks = []
        self._path = None
        self._project_modules = OrderedDict()
        self._project_recipes = OrderedDict()
        self._project_resources = OrderedDict()
        self._build_path = None
        self._workspace_name = None

    def _add_project_module(self, project, src):
        modules = self._project_modules.get(project, [])
        modules.append(src)
        self._project_modules[project] = modules

    def _get_project_modules(self, project):
        return self._project_modules.get(project, [])

    def _add_project_recipe(self, project, joltdir, src):
        recipes = self._project_recipes.get(project, [])
        recipes.append((joltdir, src))
        self._project_recipes[project] = recipes

    def _get_project_recipes(self, project):
        return self._project_recipes.get(project, [])

    def _get_first_recipe_path(self):
        for recipe in self.recipes:
            return recipe.path
        for project, recipes in self._project_recipes.items():
            for joltdir, src in recipes:
                return fs.path.join(self._path, src)
        return None

    def _add_project_resource(self, project, resource_name, resource_task):
        class ProjectResource(Alias):
            name = project + "/" + resource_name
            requires = [resource_task]

        self._tasks.append(ProjectResource)
        resources = self._project_resources.get(project, [])
        resources.append((resource_name, resource_task))
        self._project_resources[project] = resources

    def _get_project_resources(self, project):
        return self._project_resources.get(project, [])

    def _load_project_recipes(self):
        for project, recipes in self._project_recipes.items():
            resources = [resource for _, resource in self._get_project_resources(project)]
            for joltdir, src in recipes:
                joltdir = fs.path.join(self.joltdir, joltdir) if joltdir else self.joltdir
                recipe = NativeRecipe(fs.path.join(self._path, src), joltdir, project)
                recipe.load()
                for task in recipe.tasks:
                    task._resources = resources
                    attributes.requires("_resources")(task)
                self._tasks += recipe.tasks

    def _find_workspace_path(self, searchdir):
        for factory in _loaders:
            loader = factory().create(searchdir)
            if loader.recipes:
                return searchdir

        parentdir = os.path.dirname(searchdir)
        if searchdir == parentdir:
            return os.getcwd()

        return self._find_workspace_path(parentdir)

    def _get_searchpaths(self):
        return [self.workspace_path]

    def load(self, manifest=None):
        if not self.workspace_path:
            self.set_workspace_path(self._find_workspace_path(os.getcwd()))

        if not self.workspace_path:
            return []

        for searchpath in self._get_searchpaths():
            for factory in _loaders:
                loader = factory().create(searchpath)
                for recipe in loader.recipes:
                    recipe.load()
                    self._recipes.append(recipe)
                    self._tasks += recipe.tasks

        self._load_project_recipes()

        # Create workspace lock on the first loaded recipe
        if not self._lock:
            path = self._get_first_recipe_path()
            if path:
                self._lock = fasteners.InterProcessLock(path)

        return self._tasks

    def load_file(self, path, joltdir=None):
        for factory in _loaders:
            loader = factory().create(path)
            for recipe in loader.recipes:
                joltdir = fs.path.join(self.joltdir, joltdir) if joltdir else self.joltdir
                recipe.load(joltdir=joltdir)
                self._recipes.append(recipe)
                self._tasks += recipe.tasks

    def load_plugin(self, filepath):
        plugin, ext = os.path.splitext(fs.path.basename(filepath))
        loader = SourceFileLoader("jolt.plugins." + plugin, filepath)
        module = ModuleType(loader.name)
        module.__file__ = filepath
        loader.exec_module(module)
        sys.modules[loader.name] = module

    def load_plugins(self):
        searchpath = config.get("jolt", "pluginpath")
        searchpath = searchpath.split(":") if searchpath else []
        searchpath.append(fs.path.join(fs.path.dirname(__file__), "plugins"))

        for plugin in config.plugins():
            for path in searchpath:
                if "jolt.plugins." + plugin not in sys.modules:
                    module = fs.path.join(fs.path.dirname(__file__), path, plugin + ".py")
                    if fs.path.exists(module):
                        self.load_plugin(module)
                        continue
                    module = fs.path.join(fs.path.dirname(__file__), path, plugin, "__init__.py")
                    if fs.path.exists(module):
                        self.load_plugin(module)
                        continue

    @property
    def projects(self):
        return self._recipes

    @property
    def recipes(self):
        return self._recipes

    @property
    def tasks(self):
        return self._tasks

    @property
    def joltdir(self):
        return self._path

    @property
    def workspace_name(self):
        return self._workspace_name or os.path.basename(self.workspace_path)

    def set_workspace_name(self, name):
        self._workspace_name = name

    @property
    def workspace_path(self):
        return self._path

    @contextmanager
    @utils.locked
    def workspace_lock(self):
        if not self._lock:
            yield
        with self._lock:
            yield

    def set_workspace_path(self, path):
        if not self._path or len(path) < len(self._path):
            self._path = os.path.normpath(path) if path is not None else None

    @property
    def build_path(self):
        return self._build_path or os.path.join(self.workspace_path, "build")

    @property
    def build_path_rel(self):
        return os.path.relpath(self.build_path, self.workspace_path)

    def set_build_path(self, path):
        self._build_path = os.path.normpath(os.path.join(self.workspace_path, path))
        log.debug("Jolt build path: {}", self._build_path)


class RecipeExtension(ManifestExtension):
    def export_manifest(self, manifest, tasks):
        loader = JoltLoader.get()

        for recipe in loader.recipes:
            manifest_recipe = manifest.create_recipe()
            manifest_recipe.path = recipe.basepath
            manifest_recipe.source = recipe.source

        projects = set([task.task.joltproject for task in tasks])
        for project in filter(lambda x: x is not None, projects):
            manifest_project = manifest.create_project()
            manifest_project.name = project

            for name, resource_task in loader._get_project_resources(project):
                resource = manifest_project.create_resource()
                resource.name = name
                resource.text = resource_task

            for joltdir, src in loader._get_project_recipes(project):
                recipe = manifest_project.create_recipe()
                recipe.src = src
                if joltdir:
                    recipe.joltdir = joltdir

            for path in loader._get_project_modules(project):
                module = manifest_project.create_module()
                module.path = path

    def import_manifest(self, manifest):
        loader = JoltLoader.get()
        loader.set_workspace_path(manifest.get_workspace_path() or os.getcwd())
        loader.set_workspace_name(manifest.get_workspace_name())
        if manifest.build:
            loader.set_build_path(manifest.build)

        for recipe in manifest.recipes:
            recipe = Recipe(recipe.path, source=recipe.source)
            recipe.save()

        for project in manifest.projects:
            for recipe in project.recipes:
                loader._add_project_recipe(project.name, recipe.joltdir, recipe.src)

            for resource in project.resources:
                loader._add_project_resource(project.name, resource.name, resource.text)

                # Acquire resource immediately
                task = TaskRegistry.get().get_task(resource.text, manifest=manifest)
                raise_task_error_if(not isinstance(task, WorkspaceResource), task,
                                    "only workspace resources are allowed in manifest")
                task.acquire_ws()

            for module in project.modules:
                loader._add_project_module(project.name, module.src)
                sys.path.append(fs.path.join(manifest.get_workspace_path(), module.src))

    def import_protobuf(self, buildenv):
        loader = JoltLoader.get()
        loader.set_workspace_path(os.getcwd())
        loader.set_workspace_name(buildenv.workspace.name)
        if buildenv.workspace.builddir:
            loader.set_build_path(buildenv.workspace.builddir)

        # Write .jolt files into workspace
        for file in buildenv.workspace.files:
            if file.content:
                with open(file.path, "w") as f:
                    f.write(file.content)

        for project in buildenv.workspace.projects:
            for recipe in project.recipes:
                loader._add_project_recipe(project.name, recipe.workdir, recipe.path)

            for resource in project.resources:
                loader._add_project_resource(project.name, resource.alias, resource.name)

                # Acquire resource immediately
                task = TaskRegistry.get().get_task(resource.name, buildenv=buildenv)
                raise_task_error_if(
                    not isinstance(task, WorkspaceResource), task,
                    "only workspace resources are allowed in manifest")
                task.acquire_ws()

            for path in project.paths:
                loader._add_project_module(project.name, path.path)
                sys.path.append(fs.path.join(loader.workspace_path, path.path))


ManifestExtensionRegistry.add(RecipeExtension())


def get_workspacedir():
    workspacedir = JoltLoader.get().workspace_path
    assert workspacedir is not None, "No workspace present"
    return workspacedir


@contextmanager
def workspace_lock():
    with JoltLoader.get().workspace_lock():
        yield


def workspace_locked(func):
    def wrapper(*args, **kwargs):
        with workspace_lock():
            return func(*args, **kwargs)
    return wrapper


@workspace_locked
def export_workspace(tasks=None):
    loader = JoltLoader.get()
    tools = Tools()
    tree = None

    fstree_enabled = config.getboolean("jolt", "fstree", True)
    if fstree_enabled:
        fstree = tools.which("fstree")
        if not fstree:
            host = platform.system().lower()
            arch = platform.machine().lower()
            fstree = os.path.join(os.path.dirname(__file__), "bin", f"fstree-{host}-{arch}")
        if not os.path.exists(fstree):
            fstree_enabled = False
            log.warning("fstree executable not found, will not push workspace to remote cache")

    cache_grpc_uri = config.geturi("cache", "grpc_uri")
    if fstree_enabled:
        if not cache_grpc_uri:
            log.warning("No cache gRPC URI configured, will not push workspace to remote cache")
        else:
            raise_error_if(cache_grpc_uri.scheme not in ["tcp"], "Invalid scheme in cache gRPC URI config: {}", cache_grpc_uri.scheme)
            raise_error_if(not cache_grpc_uri.netloc, "Invalid network address in cache gRPC URI config: {}", cache_grpc_uri.netloc)

    # Push workspace to remote cache if possible
    if fstree_enabled and fstree and cache_grpc_uri:
        with tools.cwd(loader.workspace_path):
            cwd = tools.getcwd()
            cachedir = config.get_cachedir()
            indexhash = utils.sha1(cwd)
            indexfile = tools.expand_path(
                "{cachedir}/indexes/{}/{}",
                indexhash[:2], indexhash[2:],
                cwd=fs.posixpath.abspath(cwd),
                cachedir=cachedir)

            if not os.path.exists(indexfile):
                process = None
                try:
                    with log.progress("Indexing workspace for the first time", count=None, unit="objects", estimates=False) as progress:
                        process = subprocess.Popen(
                            [fstree, "write-tree", "--json", "--cache", cachedir, "--ignore", ".joltignore", "--index", indexfile, "--remote", cache_grpc_uri.geturl(), "--threads", str(tools.thread_count())],
                            stdout=subprocess.DEVNULL,
                            stderr=subprocess.PIPE,
                            cwd=tools.getcwd())

                        for line in iter(process.stderr.readline, b''):
                            try:
                                event = json.loads(line.decode())
                            except json.JSONDecodeError as exc:
                                log.error("Failed to decode fstree event: {}", line.decode())
                                raise exc

                            if event.get("type") in ["cache::add"]:
                                progress.update(1)

                except Exception as exc:
                    if process:
                        process.terminate()
                    raise exc
                finally:
                    if process:
                        process.wait()
                        process.stderr.close()
                        raise_error_if(process.returncode != 0, "Failed to index workspace")

            process = None
            try:
                with log.progress("Pushing workspace to remote cache", count=None, unit="objects", estimates=False) as progress:
                    process = subprocess.Popen(
                        [fstree, "write-tree-push", "--json", "--cache", cachedir, "--ignore", ".joltignore", "--index", indexfile, "--remote", cache_grpc_uri.geturl(), "--threads", str(tools.thread_count())],
                        stdout=subprocess.DEVNULL,
                        stderr=subprocess.PIPE,
                        cwd=tools.getcwd())

                    for line in iter(process.stderr.readline, b''):
                        try:
                            event = json.loads(line.decode())
                        except json.JSONDecodeError as exc:
                            log.error("Failed to decode fstree event: {}", line.decode())
                            raise exc

                        if event.get("type") in ["cache::push"]:
                            tree = event.get("path")
                            log.info("Workspace tree: {} ({} objects)", tree, event.get("value"))

                        if event.get("type") in ["cache::remote_missing_object", "cache::remote_missing_tree"]:
                            progress.refresh()

                        if event.get("type") in ["cache::push_object", "cache::push_tree"]:
                            progress.update(1)

            except Exception as exc:
                if process:
                    process.terminate()
                raise exc

            finally:
                if process:
                    process.wait()
                    process.stderr.close()
                    raise_error_if(process.returncode != 0, "Failed to push workspace to remote cache")

    workspace = common_pb.Workspace(
        builddir=loader.build_path_rel,
        cachedir=config.get_cachedir(),
        rootdir=loader.workspace_path,
        name=loader.workspace_name,
        tree=tree,
    )

    for recipe in loader.recipes:
        workspace.files.append(
            common_pb.File(
                path=recipe.basepath,
                content=recipe.source,
            )
        )

    if tasks is None:
        projects = loader._project_recipes.keys()
    else:
        projects = set([task.task.joltproject for task in tasks])

    for project in filter(lambda x: x is not None, projects):
        pb_project = common_pb.Project()
        pb_project.name = project

        for name, resource_task in loader._get_project_resources(project):
            resource = common_pb.Project.Resource()
            resource.alias = name
            resource.name = resource_task
            pb_project.resources.append(resource)

        for joltdir, src in loader._get_project_recipes(project):
            recipe = common_pb.Project.Recipe()
            recipe.path = src
            if joltdir:
                recipe.workdir = joltdir
            pb_project.recipes.append(recipe)

        for path in loader._get_project_modules(project):
            syspath = common_pb.Project.SystemPath(path=path)
            pb_project.paths.append(syspath)

        workspace.projects.append(pb_project)

    return workspace



import click
import json
import os
import sys

from jolt import cache
from jolt import cli
from jolt import config
from jolt import filesystem as fs
from jolt import graph
from jolt import log
from jolt import scheduler
from jolt import utils
from jolt import loader
from jolt.hooks import TaskHook, TaskHookFactory, TaskHookRegistry
from jolt.influence import StringInfluence
from jolt.options import JoltOptions
from jolt.plugins import ninja
from jolt.tasks import TaskRegistry
from jolt.tasks import WorkspaceResource

log.verbose("[NinjaCompDB] Loaded")


def joltdir():
    return loader.JoltLoader.get().joltdir


def patch(command, attrib, search, replace):
    command[attrib] = command[attrib].replace(search, replace)


class CompDB(object):
    def __init__(self, path="compdb/compile_commands.json", artifact=None):
        self.commands = []
        self.attribs = {}
        if artifact:
            self.path = fs.path.join(artifact.path, path)
        else:
            self.path = path

    def metapath(self, path):
        return fs.path.join(fs.path.dirname(path), ".meta.compile_commands.json")

    def read(self, path=None):
        try:
            with open(path or self.path) as f:
                self.commands = json.load(f)
            try:
                with open(self.metapath(path or self.path)) as f:
                    self.attribs = json.load(f)
            except Exception:
                self.attribs = {}
        except OSError:
            pass

    def write(self, path=None, force=False):
        if not force and not self.commands:
            return
        with open(path or self.path, "w") as f:
            json.dump(self.commands, f, indent=2)
        with open(self.metapath(path or self.path), "w") as f:
            json.dump(self.attribs, f, indent=2)

    def annotate(self, task):
        self.attribs["joltdir"] = joltdir()
        self.attribs["cachedir"] = config.get_cachedir()

    def relocate(self, task, sandboxes=False):
        for command in self.commands:
            utils.call_and_catch(patch, command, "command", self.attribs.get("joltdir", joltdir()), joltdir())
            utils.call_and_catch(patch, command, "command", self.attribs.get("cachedir", config.get_cachedir()), config.get_cachedir())
            if sandboxes:
                utils.call_and_catch(patch, command, "command", "sandboxes/", "sandboxes-reflected/")
            utils.call_and_catch(patch, command, "directory", self.attribs.get("joltdir", joltdir()), joltdir())

    def remove_wrappers(self, task):
        for command in self.commands:
            for wrapper in [task.tools.getenv("CXXWRAP", ""), task.tools.getenv("CCWRAP", "")]:
                if command["command"].startswith(wrapper):
                    command["command"] = command["command"][len(wrapper):]

    def merge(self, db):
        self.commands.extend(db.commands)


def has_incpaths(artifact):
    return artifact.cxxinfo.incpaths.count() > 0


def stage_artifacts(artifacts, tools):
    for artifact in filter(has_incpaths, artifacts):
        tools.sandbox(artifact, incremental=True)
        if fs.has_symlinks:
            tools.sandbox(artifact, incremental=True, reflect=True)


def get_compdb_artifacts(task):
    artifacts = []

    def select_compdb(artifacts):
        return list(filter(lambda a: a.name == "compdb", artifacts))

    for dep in task.children:
        artifacts.extend(dep.artifacts)
    return select_compdb(task.artifacts)[0], select_compdb(artifacts)


def get_task_artifacts(task):
    compdb_artifact = task.get_artifact("compdb")
    artifacts = task.artifacts
    for dep in task.children:
        artifacts.extend(dep.artifacts)
    return compdb_artifact, artifacts


class CompDBHooks(TaskHook):
    def __init__(self):
        self._depfiles = config.getboolean("ninja-compdb", "depfiles", True)

    def publish_compdb(self, artifact, tools):
        with tools.cwd(self.outdir):
            artifact.collect("**/*compile_commands.json", "compdb/")

    def publish_depfiles(self, artifact, tools):
        with tools.cwd(self.outdir):
            artifact.collect("**/*.d", "depfiles/")

    def task_created(self, task):
        task.task.influence.append(StringInfluence("NinjaCompDB: v3"))
        if isinstance(task.task, ninja.CXXProject):
            compdb_artifact = task.cache.get_artifact(task, "compdb")
            task.add_artifact(compdb_artifact)
            publish_compdb = getattr(task.task, "publish_compdb", None)
            if not publish_compdb:
                setattr(task.task, "publish_compdb", CompDBHooks.publish_compdb.__get__(task.task, task.task.__class__))
            if self._depfiles:
                depfile_artifact = task.cache.get_artifact(task, "depfiles")
                task.add_artifact(depfile_artifact)
                publish_depfiles = getattr(task.task, "publish_depfiles", None)
                if not publish_depfiles:
                    setattr(task.task, "publish_depfiles", CompDBHooks.publish_depfiles.__get__(task.task, task.task.__class__))

    def task_postrun(self, task, deps, tools):
        if not isinstance(task.task, ninja.CXXProject):
            return
        with tools.cwd(tools.wsroot):
            utils.call_and_catch(tools.run, "ninja -f {outdir}/build.ninja -t compdb > {outdir}/compile_commands.json")

    def task_postpublish(self, task, artifact, tools):
        if artifact.name != "compdb":
            return

        # Add information about the workspace and cachedir roots
        db = CompDB(artifact=artifact)
        db.read()
        db.annotate(task)
        db.remove_wrappers(task)
        db.write()

        if isinstance(task.task, ninja.CXXProject):
            dbpath = fs.path.join(task.task.outdir, "all_compile_commands.json")
            _, deps = get_compdb_artifacts(task)
            db = CompDB(dbpath)
            for dep in [artifact] + deps:
                depdb = CompDB(artifact=dep)
                depdb.read()
                depdb.relocate(task)
                db.merge(depdb)
            db.write()
            artifact.collect(dbpath, "compdb/", flatten=True)

    def task_finished_execution(self, task):
        if task.options.network or task.options.worker:
            return
        if not task.is_goal():
            return
        if isinstance(task.task, ninja.CXXProject):
            artifact, deps = get_compdb_artifacts(task)
            db = CompDB("compdb/all_compile_commands.json", artifact)
            db.read()
            db.relocate(task)
            outdir = task.tools.builddir("compdb", incremental=True)
            dbpath = fs.path.join(outdir, "all_compile_commands.json")
            db.write(dbpath, force=True)
            stage_artifacts(deps + [artifact], task.tools)

    def task_finished_download(self, task):
        self.task_finished_execution(task)


@TaskHookFactory.register
class CompDBHookFactory(TaskHookFactory):
    def create(self, env):
        return CompDBHooks()


@cli.cli.command(name="compdb")
@click.argument("task", type=str, nargs=-1, required=False, shell_complete=cli._autocomplete_tasks)
@click.option("-d", "--default", type=str, multiple=True, help="Override default parameter values.")
@click.pass_context
def compdb(ctx, task, default):
    """
    Generate a compilation database for a task.

    Aggregates compilation databases found in artifacts of the specified task and
    its dependencies. The commands are then post-processed and localized to the
    current workspace.

    All task artifacts are sandboxed and their directory trees are recreated
    using symlinks pointing to the origin of collected files. When opening a
    file, an IDE can then follow the symlinks into the workspace instead of
    opening files in the artifact cache.

    The database must be regenerated if dependencies or the directory tree
    of an artifact change.

    """

    manifest = ctx.obj["manifest"]
    options = JoltOptions(default=default)
    acache = cache.ArtifactCache.get(options)
    TaskHookRegistry.get(options)
    executors = scheduler.ExecutorRegistry.get(options)
    registry = TaskRegistry.get()
    strategy = scheduler.DownloadStrategy(executors, acache)
    queue = scheduler.TaskQueue(strategy, acache, {})

    for params in default:
        registry.set_default_parameters(params)

    gb = graph.GraphBuilder(registry, acache, manifest, options, progress=True)
    dag = gb.build(task)

    try:
        with log.progress("Progress", dag.number_of_tasks(), " tasks", estimates=False, debug=False) as progress:
            while dag.has_tasks() or not queue.empty():
                leafs = dag.select(lambda graph, task: task.is_ready())

                # Order the tasks by their weights to improve build times
                leafs.sort(key=lambda x: x.weight)

                while leafs:
                    task = leafs.pop()
                    queue.submit(task)

                task, error = queue.wait()

                # Materialize workspace resources so that
                # source code is available to the debugger.
                if isinstance(task.task, WorkspaceResource):
                    task.task.acquire_ws()

                # Unpack the task if it is not a resource task and has a custom unpack method
                if not task.is_resource():
                    if task.is_unpackable():
                        task.unpack()

                progress.update(1)

    except KeyboardInterrupt:
        print()
        log.warning("Interrupted by user")
        try:
            queue.abort()
            sys.exit(1)
        except KeyboardInterrupt:
            print()
            log.warning("Interrupted again, exiting")
            os._exit(1)

    finally:
        queue.shutdown()

    for goal in dag.goals:
        if goal.is_resource():
            continue
        if goal.is_alias():
            continue
        compdb_artifact, artifacts = get_task_artifacts(goal)
        db = CompDB("compdb/all_compile_commands.json", compdb_artifact)
        db.read()
        db.relocate(goal, sandboxes=fs.has_symlinks())
        outdir = goal.tools.builddir("compdb", incremental=True)
        dbpath = fs.path.join(outdir, "all_compile_commands.json")
        db.write(dbpath, force=True)
        stage_artifacts(artifacts, goal.tools)
        log.info("Compilation DB: {}", dbpath)



from jolt import filesystem as fs
from jolt import utils
from jolt.cache import ArtifactStringAttribute
from jolt.cache import ArtifactAttributeSet
from jolt.cache import ArtifactAttributeSetProvider


class EnvironmentVariable(ArtifactStringAttribute):
    def __init__(self, artifact, name):
        super(EnvironmentVariable, self).__init__(artifact, name)
        self._old_value = None

    def apply(self, task, artifact):
        self._old_value = task.tools.getenv(self.get_name())
        task.tools.setenv(self.get_name(), self.get_value())

    def unapply(self, task, artifact):
        if self._old_value:
            task.tools.setenv(self.get_name(), self._old_value)
        else:
            task.tools.setenv(self.get_name())
        self._old_value = None

    def visit(self, task, artifact, visitor):
        visitor.setenv(self.get_name(), self.get_value())


class PathEnvironmentVariable(EnvironmentVariable):
    def __init__(self, artifact, name="PATH"):
        super(PathEnvironmentVariable, self).__init__(artifact, name)

    def set_value(self, value, expand=True):
        values = utils.as_list(value)
        super(PathEnvironmentVariable, self).set_value(fs.pathsep.join(values), expand)

    def append(self, value):
        if self.get_value():
            self.set_value(self.get_value() + fs.pathsep + value)
        else:
            self.set_value(value)

    def apply(self, task, artifact):
        self._old_value = task.tools.getenv(self.get_name())
        paths = self.get_value().split(fs.pathsep)
        paths = [fs.path.join(artifact.path, path) for path in paths]
        new_val = fs.pathsep.join(paths)
        if self._old_value:
            new_val = new_val + fs.pathsep + task.tools.getenv(self.get_name())
        task.tools.setenv(self.get_name(), new_val)

    def visit(self, task, artifact, visitor):
        paths = self.get_value().split(fs.pathsep)
        paths = [fs.path.join(artifact.path, path) for path in paths]
        visitor.setenv(self.get_name(), paths)


class EnvironmentVariableSet(ArtifactAttributeSet):
    def __init__(self, artifact):
        super(EnvironmentVariableSet, self).__init__()
        super(ArtifactAttributeSet, self).__setattr__("_artifact", artifact)

    def create(self, name):
        if "PATH" in name:
            return PathEnvironmentVariable(self._artifact, name)
        return EnvironmentVariable(self._artifact, name)


@ArtifactAttributeSetProvider.Register
class EnvironmentVariableSetProvider(ArtifactAttributeSetProvider):
    def create(self, artifact):
        setattr(artifact, "environ", EnvironmentVariableSet(artifact))

    def parse(self, artifact, content):
        if "environ" not in content:
            return

        for key, value in content["environ"].items():
            getattr(artifact.environ, key).set_value(value, expand=False)

    def format(self, artifact, content):
        if "environ" not in content:
            content["environ"] = {}

        for key, attrib in artifact.environ.items():
            content["environ"][key] = attrib.get_value()

    def apply(self, task, artifact):
        artifact.environ.apply(task, artifact)

    def unapply(self, task, artifact):
        artifact.environ.unapply(task, artifact)

    def visit(self, task, artifact, visitor):
        artifact.environ.visit(task, artifact, visitor)



from jolt import config
from jolt import error
from jolt import log
from jolt.hooks import TaskHookFactory
from jolt.plugins import telemetry


log.verbose("[Dashboard] Loaded")


class DashboardHooks(telemetry.TelemetryHooks):
    def __init__(self, uri=None):
        uri = config.get("dashboard", "uri", "http://dashboard.")
        error.raise_error_if(not uri, "dashboard.uri not configured")
        super().__init__(plugin="dashboard", uri=uri + "/api/v1/tasks", local=False)


# After logstash
@TaskHookFactory.register_with_prio(20)
class DashboardFactory(TaskHookFactory):
    def create(self, env):
        return DashboardHooks()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: jolt/plugins/remote_execution/administration.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from jolt import common_pb2 as jolt_dot_common__pb2
from jolt.plugins.remote_execution import scheduler_pb2 as jolt_dot_plugins_dot_remote__execution_dot_scheduler__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n2jolt/plugins/remote_execution/administration.proto\x1a\x1bgoogle/protobuf/empty.proto\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x11jolt/common.proto\x1a-jolt/plugins/remote_execution/scheduler.proto\"\"\n\x11ListBuildsRequest\x12\r\n\x05tasks\x18\x01 \x01(\x08\"\xf4\x02\n\x12ListBuildsResponse\x12)\n\x06\x62uilds\x18\x01 \x03(\x0b\x32\x19.ListBuildsResponse.Build\x1aR\n\x04Task\x12\n\n\x02id\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x1b\n\x06status\x18\x03 \x01(\x0e\x32\x0b.TaskStatus\x12\x13\n\x0bhasObserver\x18\x04 \x01(\x08\x1a\xde\x01\n\x05\x42uild\x12\n\n\x02id\x18\x01 \x01(\t\x12\x1c\n\x06status\x18\x02 \x01(\x0e\x32\x0c.BuildStatus\x12\'\n\x05tasks\x18\x03 \x03(\x0b\x32\x18.ListBuildsResponse.Task\x12/\n\x0bscheduledAt\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x13\n\x0bhasObserver\x18\x05 \x01(\x08\x12\x16\n\x0ehasRunningTask\x18\x06 \x01(\x08\x12\x15\n\rhasQueuedTask\x18\x07 \x01(\x08\x12\r\n\x05ready\x18\x08 \x01(\x08\"\x14\n\x12ListWorkersRequest\"\x15\n\x13ListWorkersResponse2\xf9\x01\n\x0e\x41\x64ministration\x12\x38\n\x0b\x43\x61ncelBuild\x12\x13.CancelBuildRequest\x1a\x14.CancelBuildResponse\x12\x35\n\nListBuilds\x12\x12.ListBuildsRequest\x1a\x13.ListBuildsResponse\x12\x38\n\x0bListWorkers\x12\x13.ListWorkersRequest\x1a\x14.ListWorkersResponse\x12<\n\nReschedule\x12\x16.google.protobuf.Empty\x1a\x16.google.protobuf.EmptyB\x0eZ\x0cpkg/protocolb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'jolt.plugins.remote_execution.administration_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'Z\014pkg/protocol'
  _globals['_LISTBUILDSREQUEST']._serialized_start=182
  _globals['_LISTBUILDSREQUEST']._serialized_end=216
  _globals['_LISTBUILDSRESPONSE']._serialized_start=219
  _globals['_LISTBUILDSRESPONSE']._serialized_end=591
  _globals['_LISTBUILDSRESPONSE_TASK']._serialized_start=284
  _globals['_LISTBUILDSRESPONSE_TASK']._serialized_end=366
  _globals['_LISTBUILDSRESPONSE_BUILD']._serialized_start=369
  _globals['_LISTBUILDSRESPONSE_BUILD']._serialized_end=591
  _globals['_LISTWORKERSREQUEST']._serialized_start=593
  _globals['_LISTWORKERSREQUEST']._serialized_end=613
  _globals['_LISTWORKERSRESPONSE']._serialized_start=615
  _globals['_LISTWORKERSRESPONSE']._serialized_end=636
  _globals['_ADMINISTRATION']._serialized_start=639
  _globals['_ADMINISTRATION']._serialized_end=888
# @@protoc_insertion_point(module_scope)



#!/usr/bin/python
import os
import sys

from jolt import cli
from jolt import error
from jolt import log


def start_pdb(sig, frame):
    import pdb
    pdb.Pdb().set_trace(frame)


def dump_threads(sig, frame):
    import traceback
    print("\n===============================================================================")
    for threadId, stack in sys._current_frames().items():
        print("\n--- ThreadID:", threadId)
        for filename, lineno, name, line in traceback.extract_stack(stack):
            print('File: "{}", line {}, in {}'.format(filename, lineno, name))
            if line:
                print("  " + line.strip())
    print("\n===============================================================================\n")


def main():
    if os.name == "posix":
        import signal
        signal.signal(signal.SIGUSR1, start_pdb)
        signal.signal(signal.SIGUSR2, dump_threads)

    try:
        cli.cli(obj=dict())
    except KeyboardInterrupt:
        log.warning("Interrupted by user")
        if cli.debug_enabled:
            import pdb
            extype, value, tb = sys.exc_info()
            pdb.post_mortem(tb)
        sys.exit(1)
    except error.LoggedJoltError as e:
        log.error(log.format_exception_msg(e.exc))
        if cli.debug_enabled:
            import pdb
            extype, value, tb = sys.exc_info()
            pdb.post_mortem(tb)
        sys.exit(1)
    except Exception as e:
        log.exception(e, error=True)
        if cli.debug_enabled:
            import pdb
            extype, value, tb = sys.exc_info()
            pdb.post_mortem(tb)
        sys.exit(1)


if __name__ == "__main__":
    main()




from jolt import config
from jolt import filesystem as fs
from jolt import loader
from jolt import log
from jolt import utils
from jolt.error import raise_error_if
from jolt.hooks import CliHook, CliHookFactory
from jolt.hooks import TaskHook, TaskHookFactory

from contextlib import contextmanager

log.verbose("[Symlinks] Loaded")


class SymlinkTaskHooks(TaskHook):
    def __init__(self):
        self._path = config.get("symlinks", "path", "artifacts")
        raise_error_if(not self._path, "symlinks.path not configured")

    @property
    def rootpath(self):
        return fs.path.normpath(
            fs.path.join(
                fs.path.dirname(loader.JoltLoader.get().build_path),
                self._path
            )
        )

    def task_finished(self, task):
        if not task.has_artifact():
            return

        for artifact in task.artifacts:
            srcpath = artifact.final_path
            if artifact.name == "main":
                destpath = fs.path.join(
                    self.rootpath,
                    utils.canonical(task.short_qualified_name),
                )
            else:
                destpath = fs.path.join(
                    self.rootpath,
                    artifact.name + "@" + utils.canonical(task.short_qualified_name),
                )

            if fs.path.exists(srcpath):
                fs.unlink(destpath, ignore_errors=True)
                fs.makedirs(fs.path.dirname(destpath))
                fs.symlink(srcpath, destpath)

    def task_pruned(self, task):
        self.task_finished(task)

    def task_skipped(self, task):
        self.task_finished(task)


class SymlinkCliHooks(CliHook):
    def __init__(self):
        self._path = config.get("symlinks", "path", "artifacts")
        raise_error_if(not self._path, "symlinks.path not configured")

    @contextmanager
    def cli_clean(self, *args, **kwargs):
        yield
        try:
            path = fs.path.join(loader.get_workspacedir(), self._path)
            fs.rmtree(path, onerror=fs.onerror_warning)
        except (AssertionError, FileNotFoundError):
            pass


@TaskHookFactory.register
class SymlinkTaskHookFactory(TaskHookFactory):
    def create(self, env):
        return SymlinkTaskHooks()


@CliHookFactory.register
class SymlinkCliHookFactory(CliHookFactory):
    def create(self, env):
        return SymlinkCliHooks()



import os
import pygit2
import re
import urllib.parse

from jolt.tasks import BooleanParameter, Export, Parameter, TaskRegistry, WorkspaceResource
from jolt.influence import FileInfluence, HashInfluenceRegistry
from jolt.tools import Tools
from jolt.loader import JoltLoader, workspace_locked
from jolt import config
from jolt import filesystem as fs
from jolt import log
from jolt import utils
from jolt.error import JoltCommandError
from jolt.error import raise_error
from jolt.error import raise_error_if
from jolt.error import raise_task_error_if


log.verbose("[Git] Loaded")


class GitRepository(object):
    def __init__(self, url, path, relpath, refspecs=None):
        self.path = path
        self.relpath = relpath
        self.gitpath = None
        self.tools = Tools()
        self.url = url
        self.default_refspecs = [
            '+refs/heads/*:refs/remotes/origin/*',
            '+refs/tags/*:refs/remotes/origin/*',
        ]
        self.refspecs = refspecs or []
        self._tree_hash = {}
        self._original_head = True
        self._last_rev = None
        self._init_repo()

    def _init_repo(self):
        gitpath = os.path.join(self.path, ".git")
        if os.path.isdir(os.path.join(self.path, ".git")):
            self.gitpath = gitpath
            self.repository = pygit2.Repository(self.gitpath)
        elif os.path.exists(gitpath):
            with self.tools.cwd(self.path):
                path = self.tools.run("git rev-parse --git-dir", output=False)
            self.gitpath = os.path.join(self.path, path)
            self.repository = pygit2.Repository(self.gitpath)
        else:
            self.repository = None

    @utils.cached.instance
    def _git_folder(self):
        return self.gitpath

    @utils.cached.instance
    def _git_index(self):
        return fs.path.join(self.gitpath, "index")

    @utils.cached.instance
    def _git_jolt_index(self):
        return fs.path.join(self.gitpath, "jolt-index")

    def is_cloned(self):
        return self.gitpath is not None

    def is_indexed(self):
        return self.is_cloned() and fs.path.exists(self._git_index())

    def clone(self):
        log.info("Cloning into {0}", self.path)

        # Get reference repository path
        refroot = config.get("git", "reference", None)
        refpath = None
        if refroot:
            # Append <host>/<path> to the reference root
            url = urllib.parse.urlparse(str(self.url))
            refpath = fs.path.join(refroot, url.hostname, url.path.lstrip("/"))
            refpath = fs.path.abspath(refpath)

        # If the directory exists, initialize the repository instead of cloning
        if fs.path.exists(self.path):
            with self.tools.cwd(self.path):
                self.tools.run("git init", output_on_error=True)

                # Set the reference repository if available
                if refpath:
                    # Check if the reference repository is a git repository
                    objpath = os.path.join(refpath, ".git", "objects")
                    objpath_bare = os.path.join(refpath, "objects")
                    if os.path.isdir(objpath):
                        refpath = objpath
                    elif os.path.isdir(objpath_bare):
                        refpath = objpath_bare
                    else:
                        refpath = None

                if refpath:
                    self.tools.mkdir(".git/objects/info")
                    self.tools.write_file(".git/objects/info/alternates", refpath)

                self.tools.run("git remote add origin {}", self.url, output_on_error=True)
                self.tools.run("git fetch origin", output_on_error=True)
                self.tools.run("git checkout -f FETCH_HEAD", output_on_error=True)
        else:
            if refpath and os.path.isdir(refpath):
                self.tools.run("git clone --reference-if-able {0} {1} {2}", refpath, self.url, self.path, output_on_error=True)
            else:
                self.tools.run("git clone  {0} {1}", self.url, self.path, output_on_error=True)

        self._init_repo()
        raise_error_if(
            self.repository is None,
            "Failed to clone repository '{0}'", self.relpath)

    @utils.cached.instance
    def diff_unchecked(self):
        if not self.is_indexed():
            return ""

        # Build the jolt index file
        self.write_tree()

        # Diff against the jolt index
        with self.tools.environ(GIT_INDEX_FILE=self._git_jolt_index()), self.tools.cwd(self.path):
            return self.tools.run("git diff --binary --no-ext-diff HEAD ./",
                                  output_on_error=True,
                                  output_rstrip=False)

    def diff(self):
        diff = self.diff_unchecked()
        dlim = config.getsize("git", "maxdiffsize", "1 MiB")
        raise_error_if(
            len(diff) > dlim,
            "Repository '{}' has uncommitted changes. Size of patch exceeds configured transfer limit ({} > {} bytes)."
            .format(self.relpath, len(diff), dlim))
        return diff

    def patch(self, patch):
        if not patch:
            return
        with self.tools.cwd(self.path), self.tools.tmpdir("git") as tmp:
            patchfile = fs.path.join(tmp, "jolt.diff")
            with open(patchfile, "wb") as f:
                f.write(patch.encode())
            log.info("Applying patch to {0}", self.path)
            self.tools.run("git apply --whitespace=nowarn {patchfile}", patchfile=patchfile)

    def head(self):
        if not self.is_cloned():
            return None
        return str(self.repository.head.target)

    def is_head(self, rev):
        return self.is_indexed() and self.head() == rev

    def is_original_head(self):
        return self._original_head

    def is_valid_sha(self, rev):
        return re.match(r"[0-9a-f]{40}", rev)

    def rev_parse(self, rev):
        if self.is_valid_sha(rev):
            return rev
        with self.tools.cwd(self.path):
            try:
                commit = self.repository.revparse_single(rev)
            except KeyError:
                self.fetch(commit=rev)
                try:
                    commit = self.repository.revparse_single(rev)
                except Exception:
                    raise_error("Invalid git reference: {}", rev)
            try:
                return str(commit.id)
            except Exception:
                return str(commit)

    @utils.cached.instance
    @workspace_locked
    def write_tree(self):
        tools = Tools()
        with tools.cwd(self._git_folder()):
            tools.copy("index", "jolt-index")
        with tools.environ(GIT_INDEX_FILE=self._git_jolt_index()), tools.cwd(self.path):
            tree = tools.run(
                "git -c core.safecrlf=false add -A && git write-tree",
                output_on_error=True)
        return tree

    def tree_hash(self, rev=None, path="/"):
        # When rev is None, the caller want the tree hash of the repository's
        # current workspace state. If no checkout has been made, that would be the
        # tree that was written upon initialization of the repository as it
        # includes any uncommitted changes. If a checkout has been made since
        # the repo was initialized, make this an explicit request for the current
        # head - there can be no local changes.
        if rev is None:
            if self.is_original_head():
                tree = self.repository.get(self.write_tree())
            else:
                rev = self.head()

        path = fs.path.normpath(path)
        full_path = fs.path.join(self.path, path) if path != "/" else self.path

        # Lookup tree hash value in cache
        value = self._tree_hash.get((full_path, rev))
        if value is not None:
            return value

        # Translate explicit rev to tree
        if rev is not None:
            commit = self.rev_parse(rev)
            obj = self.repository.get(commit)
            try:
                tree = obj.tree
            except AttributeError:
                tree = obj.get_object().tree

        # Traverse tree from root to requested path
        if path != "/":
            tree = tree[fs.as_posix(path)]

        # Update tree hash cache
        self._tree_hash[(full_path, rev)] = value = tree.id

        return value

    def clean(self):
        with self.tools.cwd(self.path):
            return self.tools.run("git clean -dfx", output_on_error=True)

    def reset(self):
        with self.tools.cwd(self.path):
            return self.tools.run("git reset --hard", output_on_error=True)

    def fetch(self, commit=None):
        if commit and not self.is_valid_sha(commit):
            commit = None

        refspec = " ".join(self.default_refspecs + self.refspecs)
        with self.tools.cwd(self.path):
            log.info("Fetching {0} from {1}", commit or refspec or 'commits', self.url)
            self.tools.run(
                "git fetch {url} {what}",
                url=self.url,
                what=commit or refspec or '',
                output_on_error=True)

    def checkout(self, rev, commit=None):
        if rev == self._last_rev:
            log.debug("Checkout skipped, already @ {}", rev)
            return False
        log.verbose("Checking out {0} in {1}", rev, self.path)
        with self.tools.cwd(self.path):
            try:
                self.tools.run("git checkout -f {rev}", rev=rev, output=False)
            except Exception:
                self.fetch(commit=commit)
                try:
                    self.tools.run("git checkout -f {rev}", rev=rev, output_on_error=True)
                except Exception:
                    raise_error("Commit does not exist in remote for '{}': {}", self.relpath, rev)
        self._original_head = False
        self._last_rev = rev
        return True


_gits = {}


def new_git(url, path, relpath, refspecs=None):
    refspecs = utils.as_list(refspecs or [])
    try:
        git = _gits[path]
        raise_error_if(git.url != url, "Multiple git repositories required at {}", relpath)
        raise_error_if(git.refspecs != refspecs,
                       "Conflicting refspecs detected for git repository at  {}", relpath)
        return git
    except Exception:
        git = _gits[path] = GitRepository(url, path, relpath, refspecs)
        return git


class GitInfluenceProvider(FileInfluence):
    name = "Git"

    def __init__(self, path):
        super().__init__(path)
        self.path = path.rstrip(fs.sep)
        self.name = GitInfluenceProvider.name

    @property
    def joltdir(self):
        return JoltLoader.get().joltdir

    def _find_dotgit(self, path):
        ppath = None
        while path != ppath:
            if fs.path.exists(fs.path.join(path, ".git")):
                return path
            ppath = path
            path = fs.path.dirname(path)
        raise_error("No git repository found at '{}'", self.path)

    @utils.cached.instance
    def get_influence(self, task):
        tools = Tools(task, task.joltdir)
        path = tools.expand_path(self.path)
        git_abs = self._find_dotgit(path)
        git_rel = git_abs[len(self.joltdir) + 1:]
        relpath = path[len(git_abs) + 1:]
        relpath = fs.as_posix(relpath) if relpath else relpath

        if not fs.path.exists(path):
            return "{0}/{1}: N/A".format(git_rel, relpath)
        try:
            git = new_git(None, git_abs, git_rel)
            return "{0}/{1}: {2}".format(git_rel, relpath, git.tree_hash(path=relpath or "/"))
        except KeyError:
            return "{0}/{1}: N/A".format(git_rel, relpath)
        except JoltCommandError as e:
            stderr = "\n".join(e.stderr)
            if "exists on disk, but not in" in stderr:
                return "{0}/{1}: N/A".format(git_rel, relpath)
            raise e

    def is_influenced_by(self, task, path):
        tools = Tools(task, task.joltdir)
        gitpath = tools.expand_path(self.path)
        return fs.is_relative_to(path, gitpath)


def global_influence(path, cls=GitInfluenceProvider):
    HashInfluenceRegistry.get().register(cls(path))


def influence(path, git_cls=GitInfluenceProvider):
    def _decorate(cls):
        _old_influence = cls._influence

        def _influence(self, *args, **kwargs):
            influence = _old_influence(self, *args, **kwargs)
            influence.append(git_cls(path=path))
            return influence

        cls._influence = _influence
        return cls
    return _decorate


class ErrorDict(dict):
    """ A dict that raises an error if the value of a key is None. """

    def __init__(self, repo):
        self.repo = repo

    def __getitem__(self, key):
        value = super().__getitem__(key)
        raise_task_error_if(value is None, self.repo, "Git repository '{0}' referenced in influence collection before being cloned/checked out. Assign hash=true to the git requirement.", key)
        return value


class Git(WorkspaceResource, FileInfluence):
    """
    Resource that clones and monitors a Git repo.

    By default, the repo is cloned into a build directory named after
    the resource. The 'path' parameter can be used to specify a different
    location relative to the workspace root.

    The path of the cloned repo is made available to consuming tasks
    through their 'git' attribute. The 'git' attribute is a dictionary
    where the key is the name of the git repository and the value is
    the relative path to the repository from the consuming task's
    workspace.

    The resource influences the hash of consuming tasks, causing tasks
    to be re-executed if the cloned repo is modified.

    The plugin must be loaded before it can be used. This is done by
    importing the module, or by adding the following line to the
    configuration file:

    .. code-block:: ini

            [git]

    Example:

    .. code-block:: python

            from jolt.plugins import git

            class Example(Task):
                requires = ["git:url=https://github.com/user/repo.git"]

                def run(self, deps, tools):
                    self.info("The git repo is located at: {git[repo]}")
                    with tools.cwd(self.git["repo"]):
                        tools.run("make")

    """
    name = "git"

    url = Parameter(help="URL to the git repo to be cloned. Required.")
    """ URL to the git repo to be cloned. Required. """

    rev = Parameter(required=False, help="Specific commit or tag to be checked out. Optional.")
    """ Specific commit or tag to be checked out. Optional. """

    hash = BooleanParameter(required=False, help="Let repo content influence the hash of consuming tasks.")
    """ Let repo content influence the hash of consuming tasks. Default ``True``. Optional. """

    path = Parameter(required=False, help="Local path where the repository should be cloned.")
    """ Alternative path where the repository should be cloned. Relative to ``joltdir``. Optional. """

    _revision = Export(value=lambda t: t._export_revision())
    """ To worker exported value of the revision to be checked out. """

    _diff = Export(value=lambda t: t.git.diff(), encoded=True)
    """ To worker exported value of the diff of the repo. """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.joltdir = JoltLoader.get().joltdir

        # Set the path to the repo
        if self.path.is_unset():
            self.abspath = self.tools.builddir(utils.canonical(self.short_qualified_name), incremental="always", unique=False)
            self.relpath = fs.path.relpath(self.abspath, self.tools.wsroot)
        else:
            self.abspath = fs.path.join(self.joltdir, str(self.path) or self._get_name())
            self.relpath = fs.path.relpath(self.abspath, self.tools.wsroot)

        # Create the git repository
        self.refspecs = kwargs.get("refspecs", [])
        self.git = new_git(self.url, self.abspath, self.relpath, self.refspecs)

    @utils.cached.instance
    def _get_name(self):
        repo = fs.path.basename(self.url.get_value())
        name, _ = fs.path.splitext(repo)
        return name

    def _export_revision(self):
        return self.rev.value or self.git.head()

    def _get_revision(self):
        if self._revision.is_imported:
            return self._revision.value
        if not self.rev.is_unset():
            return self.rev.get_value()
        return None

    def _assign_git(self, task, none=False):
        if not hasattr(task, "git"):
            task.git = ErrorDict(self)
        if none:
            # None means the git repo is not cloned or checked out
            # and should not be included in the git dictionary
            # of the consuming task yet. If the consuming task
            # requires the git repo for its influence collection,
            # the dict will raise an error. The solution is to
            # assign hash=true to the git requirement which
            # will cause the git repo to be cloned and checked out
            # before the influence collection is performed.
            task.git[self._get_name()] = None
        else:
            # Assign the git repo to the consuming task.
            # The git repo is cloned and checked out before
            # any influence collection is performed.
            task.git[self._get_name()] = fs.path.relpath(self.abspath, task.joltdir)

    def acquire(self, artifact, deps, tools, owner):
        self._acquire_ws()
        self._assign_git(owner)
        artifact.worktree = fs.path.relpath(self.abspath, owner.joltdir)

    def prepare_ws_for(self, task):
        """ Prepare the workspace for the task.

        :param task: The task to prepare the workspace for.
        """
        if not self._must_influence():
            # The content of the git repo is not required to influence the hash of the
            # consumer task. The repo is therefore not cloned or checked out
            # until the consumer is executed. Raise an error if the git repo
            # is required for the influence collection of the consumer task.
            self._assign_git(task, none=True)
            return
        # The content of the git repo is required to influence the hash of the consumer task.
        self._assign_git(task)

    def acquire_ws(self):
        """ Clone and/or checkout the git repo if required """
        if self._must_influence() or self._revision.is_imported:
            self._acquire_ws()

    def _acquire_ws(self):
        commit = None
        if not self.git.is_cloned():
            self.git.clone()
        if not self._revision.is_imported:
            self.git.diff_unchecked()
        else:
            commit = self._revision.value
        rev = self._get_revision()
        if rev is not None:
            raise_task_error_if(
                not self._revision.is_imported and not self.rev.is_unset() and self.git.diff(), self,
                "Explicit revision requested but git repo '{0}' has local changes, refusing checkout", self.git.relpath)
            # Should be safe to do this now
            rev = self.git.rev_parse(rev)
            if not self.git.is_head(rev) or self._revision.is_imported:
                if self.git.checkout(rev, commit=commit):
                    self.git.clean()
                    self.git.patch(self._diff.value)

    def _must_influence(self):
        """ Check if the git repo must influence the hash of the consumer task."""

        # If the hash parameter is set, honor it
        if self.hash.is_set():
            return self.hash

        # If the revision parameter is not set, the git repo must influence the hash
        if self.rev.is_unset():
            return True

        # If the revision parameter is set, no influence is needed since the
        # revision is fixed and repository content will not change.
        return False

    def is_influenced_by(self, task, path):
        influencing = self._must_influence() or self.rev.is_set()
        return influencing and fs.is_relative_to(path, self.abspath)

    def _influence(self):
        influence = super()._influence()
        return influence + [self] if self._must_influence() else influence

    @utils.cached.instance
    def get_influence(self, task):
        if not self.git.is_cloned():
            self.git.clone()
        if not self._revision.is_imported:
            self.git.diff_unchecked()
        rev = self._get_revision()

        try:
            if rev is not None:
                th = self.git.tree_hash(rev)
            else:
                th = self.git.tree_hash()
        except KeyError:
            th = "N/A"

        return "{0}: {1}".format(self.git.relpath, th)


TaskRegistry.get().add_task_class(Git)



from .error import JoltError
from .error import JoltCommandError
from .error import JoltTimeoutError

from .tasks import Alias
from .tasks import BooleanParameter
from .tasks import Chroot
from .tasks import Download
from .tasks import EnvironExport
from .tasks import Export
from .tasks import IntParameter
from .tasks import ListParameter
from .tasks import MultiTask
from .tasks import Parameter
from .tasks import Resource
from .tasks import Runner
from .tasks import Script
from .tasks import Task
from .tasks import TaskGenerator
from .tasks import Test
from .tasks import attributes

from .cache import Artifact
from .cache import Context

from .tools import Tools

from .version import __version__
from . import expires
from . import influence

__all__ = (
    "Alias",
    "Artifact",
    "BooleanParameter",
    "Chroot",
    "Context",
    "Download",
    "EnvironExport",
    "Export",
    "IntParameter",
    "JoltError",
    "JoltCommandError",
    "JoltTimeoutError",
    "ListParameter",
    "MultiTask",
    "Parameter",
    "Resource",
    "Runner",
    "Script",
    "Task",
    "TaskGenerator",
    "Test",
    "Tools",
    "__version__",
    "attributes",
    "expires",
    "influence",
)

name = "jolt"


def include(joltfile, joltdir=None):
    """ Include another Python file with Jolt tasks.

      :param joltfile: The path to the Jolt file to include.
      :type joltfile: str

      :param joltdir: The directory to search for Jolt files.
      :type joltdir: str

    Example:

      .. code-block:: python

            from jolt import include

            include("joltfile.py")

      """
    try:
        from os import path
        from jolt.loader import JoltLoader
        dirpath = path.dirname(joltfile)
        filepath = path.join(dirpath, joltfile)
        JoltLoader.get().load_file(filepath, joltdir=joltdir)
    except Exception as e:
        from jolt.error import raise_error
        raise_error("Failed to load '{0}': {1}", joltfile, str(e))



import json
from os import path

from jolt import Task
from jolt import influence
from jolt.error import raise_task_error_if


@influence.attribute("conanfile")
@influence.attribute("generators")
@influence.attribute("incremental")
@influence.attribute("options")
@influence.attribute("packages")
class Conan(Task):
    """
    Conan package installer task.

    This task base class can be used to fetch, build and publish Conan packages
    as Jolt artifacts. All package metadata is transfered from the Conan package
    manifest to the Jolt artifact so that no manual configuration of include
    paths, library paths, macros, etc is required.

    An existing installation of Conan is required. Please visit https://conan.io/
    for installation instructions and documentation.

    A minimal task to download and publish the Boost C++ libraries can look like this:

    .. code-block:: python

        from jolt.plugins.conan import Conan

        class Boost(Conan):
            packages = ["boost/1.74.0"]

    Boost may then be used from Ninja tasks by declaring a requirement:

    .. code-block:: python

        from jolt.plugins.ninja import CXXExecutable

        class BoostApplication(CXXExecutable):
            requires = ["boost"]
            sources = ["main.cpp"]

    The task supports using an existing conanfile.txt, but it is not required.
    Packages are installed into and collected from Jolt build directories. The
    user's regular Conan cache will not be affected.

    """

    abstract = True

    conanfile = None
    """
    An existing conanfile.txt file to use.

    Instead of generating the conanfile.txt file on-demand, an external
    file may be used. If this attribute is set, the ``generators``, ``options``
    and ``packages`` attributes must not be set.

    See Conan documentation for further details.
    """

    packages = []
    """
    A list of Conan package references to collect and publish.

    The reference format is ``PkgName/<version>@user/channel``. See Conan
    documentation for further details.

    Any {keyword} arguments, or macros, found in the strings are automatically
    expanded to the value of the associated task's parameters and properties.

    Example:

    .. code-block:: python

        sdl_version = Parameter("2.0.12")

        packages = [
            "boost/1.74.0",
            "sdl2/{sdl_version}@bincrafters/stable",
        ]

    """

    options = []
    """
    A list of Conan package options to apply

    The option format is ``PkgName:Option=Value``. See Conan
    documentation for further details.

    Any {keyword} arguments, or macros, found in the strings are automatically
    expanded to the value of the associated task's parameters and properties.

    Example:

    .. code-block:: python

        options = [
            "boost:shared=True",
            "zlib:shared=True",
        ]

    """

    settings = []
    """
    A list of Conan settings to apply

    The settings format is ``Option=Value``. See Conan
    documentation for further details.

    Any {keyword} arguments, or macros, found in the strings are automatically
    expanded to the value of the associated task's parameters and properties.

    Example:

    .. code-block:: python

        settings = [
            "compiler.libcxx=libstdc++11",
        ]

    """

    generators = []
    """
    A list of Conan generators to use.

    See Conan documentation for details about supported generators.
    The json generator is always used.

    Example:

    .. code-block:: python

        generators = ["cmake"]

    """

    remotes = {}
    """
    A dictionary with Conan remotes to use when fetching packages.

    The dictionary key is the name of remote and its value is the URL.

    Example:

    .. code-block:: python

        remotes = {
            "bincrafters": "https://api.bintray.com/conan/bincrafters/public-conan"
        }

    """

    incremental = True
    """
    Keep installed packages in the Conan cache between Jolt invokations.

    If incremental build is disabled, the Jolt Conan cache is removed
    before execution begins.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.conanfile:
            self.influence.append(influence.FileInfluence(self.conanfile))

    def _generators(self):
        return ["json"] + self.generators

    def _options(self):
        return [] + self.options

    def _settings(self):
        return [] + self.settings

    def _packages(self):
        return [] + self.packages

    def _remotes(self):
        return self.remotes

    def run(self, deps, tools):
        raise_task_error_if(
            not tools.which("conan"), self,
            "Conan: Conan is not installed in the PATH")
        raise_task_error_if(
            self.conanfile and (self._generators() or self._packages() or self._options()), self,
            "Conan: 'conanfile' attribute cannot be used with other attributes")

        conanfile = tools.expand_path(self.conanfile) if self.conanfile else None

        with tools.cwd(tools.builddir()):
            if conanfile is None or not path.exists(conanfile):
                conanfile = "conanfile.txt"
                self.info("Creating conanfile.txt")
                self.tools.write_file(conanfile, "[requires]\n")
                for pkg in self._packages():
                    self.tools.append_file(conanfile, pkg + "\n")

            with tools.environ(CONAN_USER_HOME=tools.builddir("conan", incremental=self.incremental)):
                for remote, url in self._remotes().items():
                    self.info("Registering remote '{}'", remote)
                    tools.run("conan remote add -f {} {}", remote, url, output_on_error=True)

                self.info("Installing packages into the Conan cache")
                generators = " ".join(["-g " + gen for gen in self._generators()])
                options = " ".join(["-o " + opt for opt in self._options()])
                settings = " ".join(["-s " + opt for opt in self._settings()])
                tools.run("conan install --build=missing -u -if . {} {} {} {}", generators, options, settings, conanfile)

            self.info("Parsing manifest")
            self._manifest = json.loads(tools.read_file("conanbuildinfo.json"))

            for dep in self._manifest["dependencies"]:
                self.info("Collecting '{}' files from: {}", dep["name"], dep["rootpath"])
                tools.copy(dep["rootpath"], dep["name"])

    def publish(self, artifact, tools):
        self.info("Publishing package files")
        with tools.cwd(tools.builddir()):
            artifact.collect("*")

        self.info("Publishing metadata")
        for dep in self._manifest["dependencies"]:
            for incpath in dep["include_paths"]:
                artifact.cxxinfo.incpaths.append(path.join(dep["name"], path.relpath(incpath, dep["rootpath"])))
            for libpath in dep["lib_paths"]:
                artifact.cxxinfo.libpaths.append(path.join(dep["name"], path.relpath(libpath, dep["rootpath"])))
            for binpath in dep["bin_paths"]:
                artifact.environ.PATH.append(path.join(dep["name"], path.relpath(binpath, dep["rootpath"])))
            artifact.cxxinfo.libraries.append(dep["libs"])
            artifact.cxxinfo.libraries.append(dep["system_libs"])
            artifact.cxxinfo.macros.append(dep["defines"])
            artifact.cxxinfo.cflags.append(dep["cflags"])
            artifact.cxxinfo.cxxflags.append(dep["cxxflags"])
            artifact.cxxinfo.ldflags.append(dep["exelinkflags"])



from xml.dom import minidom
from xml.etree import ElementTree as ET
import os

from jolt.xmldom import Attribute, Composition, SubElement, Element, ElementTree
from jolt import filesystem as fs
from jolt import log


@Attribute('name')
@Attribute('value', child=True)
class _JoltAttribute(SubElement):
    def __init__(self, elem=None):
        super(_JoltAttribute, self).__init__('attribute', elem=elem)


@Attribute('path')
@Attribute('source', child=True, zlib=True)
class _JoltRecipe(SubElement):
    def __init__(self, elem=None):
        super(_JoltRecipe, self).__init__('recipe', elem=elem)


@Attribute('src')
@Attribute('dest')
class _JoltProjectLink(SubElement):
    def __init__(self, elem=None):
        super(_JoltProjectLink, self).__init__('link', elem=elem)


@Attribute('src')
@Attribute('joltdir')
class _JoltProjectRecipe(SubElement):
    def __init__(self, elem=None):
        super(_JoltProjectRecipe, self).__init__('recipe', elem=elem)


@Attribute('src')
class _JoltProjectModule(SubElement):
    def __init__(self, elem=None):
        super(_JoltProjectModule, self).__init__('module', elem=elem)

    @property
    def path(self):
        return self.src

    @path.setter
    def path(self, value):
        self.src = value


@Attribute('name')
class _JoltProjectResource(SubElement):
    def __init__(self, elem=None):
        super(_JoltProjectResource, self).__init__('resource', elem=elem)


@Attribute('name')
@Composition(_JoltProjectLink, "link")
@Composition(_JoltProjectModule, "module")
@Composition(_JoltProjectRecipe, "recipe")
@Composition(_JoltProjectResource, "resource")
class _JoltProject(SubElement):
    def __init__(self, elem=None):
        super(_JoltProject, self).__init__('project', elem=elem)


@Attribute('type', child=True)
@Attribute('location', child=True)
@Attribute('message', child=True)
@Attribute('details', child=True)
class _JoltTaskError(SubElement):
    def __init__(self, elem=None):
        super(_JoltTaskError, self).__init__('error', elem=elem)


@Attribute('name')
@Attribute('duration', child=True)
@Attribute('goal', child=True)
@Attribute('identity', child=True)
@Attribute('instance', child=True)
@Attribute('logstash', child=True)
@Attribute('result', child=True)
@Composition(_JoltAttribute, "attribute")
@Composition(_JoltTaskError, "error")
class _JoltTask(SubElement):
    def __init__(self, elem=None):
        super(_JoltTask, self).__init__('task', elem=elem)


@Attribute('name')
class _JoltDefault(SubElement):
    def __init__(self, elem=None):
        super(_JoltDefault, self).__init__('default', elem=elem)


@Composition(_JoltDefault, "default")
@Composition(_JoltTask, "task")
class _JoltBuild(SubElement):
    def __init__(self, elem=None):
        super(_JoltBuild, self).__init__('build', elem=elem)


@Attribute("key")
@Attribute("value")
class _JoltNetworkParameter(SubElement):
    def __init__(self, elem=None):
        super(_JoltNetworkParameter, self).__init__('parameter', elem=elem)


@Attribute("config", child=True, zlib=True)
@Attribute("stdout", child=True, zlib=True)
@Attribute("stderr", child=True, zlib=True)
@Attribute("result", child=True)
@Attribute("duration", child=True)
@Attribute("name")
@Attribute("workspace")
@Attribute("build")
@Attribute("version")
@Composition(_JoltRecipe, "recipe")
@Composition(_JoltTask, "task")
@Composition(_JoltBuild, "build")
@Composition(_JoltNetworkParameter, "parameter")
@Composition(_JoltProject, "project")
class JoltManifest(ElementTree):
    def __init__(self):
        super(JoltManifest, self).__init__(element=Element('jolt-manifest'))
        self._identities = None
        self._elem = self.getroot()
        self.path = None

    def is_valid(self):
        return self.path is not None

    def get_workspace_path(self):
        if self.path is None:
            return None
        joltdir = fs.path.dirname(self.path)
        if self.workspace:
            joltdir = fs.path.normpath(fs.path.join(joltdir, self.workspace))
        return joltdir

    def get_workspace_name(self):
        if self.name:
            return self.name
        if self.path is None:
            return None
        return fs.path.basename(fs.path.dirname(self.path))

    @property
    def attrib(self):
        return self.getroot().attrib

    def append(self, element):
        SubElement(elem=self.getroot()).append(element)

    def get(self, key):
        return self.getroot().get(key)

    def set(self, key, value):
        self.getroot().set(key, value)

    def remove(self, child):
        self.getroot().remove(child)

    def parse(self, filename="default.joltxmanifest"):
        path = os.getcwd()
        filepath = fs.path.join(path, filename)
        while not fs.path.exists(filepath):
            opath = path
            path = fs.path.dirname(path)
            if path == opath:
                break
            filepath = fs.path.join(path, filename)
        if path == fs.sep:
            raise Exception("couldn't find manifest file")
        with open(filepath) as f:
            self.parsestring(f.read())
            log.verbose("Loaded: {0}", filepath)
            self.path = filepath
            return self
        raise Exception("failed to parse xml file")

    def parsestring(self, string):
        root = ET.fromstring(string)
        for elem in root.iter():
            if elem.text is not None:
                elem.text = elem.text.strip()
            if elem.tail is not None:
                elem.tail = elem.tail.strip()
        self._setroot(root)
        self._elem = root
        return self

    def format(self):
        return minidom.parseString(ET.tostring(self.getroot())).toprettyxml(indent="  ")

    def transform(self, xsltfile):
        from lxml import etree as lxmlET
        manifest = lxmlET.fromstring(self.format())
        xslt = lxmlET.parse(xsltfile)
        transform = lxmlET.XSLT(xslt)
        document = transform(manifest)
        return minidom.parseString(lxmlET.tostring(document)).toprettyxml(indent="  ")

    def write(self, filename):
        with open(filename, 'w') as f:
            f.write(self.format())

    def has_task(self, task):
        return self.find("./task[@identity='{}']".format(task.identity)) is not None

    def has_failure(self):
        if self.result and self.result == "FAILED":
            return True
        return self.find("./task[result='FAILED']") is not None

    def has_unstable(self):
        return self.find("./task[result='UNSTABLE']") is not None

    def has_tasks(self):
        return self.find("./task") is not None

    def find_task(self, task):
        match = self.find("./task[@name='{0}']".format(task))
        if match is None:
            return None
        return _JoltTask(elem=match)

    def get_parameter(self, key):
        match = self.find("./parameter[@key='{0}']".format(key))
        if match is None:
            return None
        return match.get("value")

    @property
    def task_identities(self):
        if self._identities is not None:
            return self._identities
        self._identities = {}
        for manifest_task in self.tasks:
            self._identities[manifest_task.name] = manifest_task.identity
        return self._identities

    @staticmethod
    def export(task):
        manifest = JoltManifest()
        ManifestExtensionRegistry.export_manifest(manifest, task)
        return manifest

    def process_import(self):
        ManifestExtensionRegistry.import_manifest(self)


class ManifestExtensionRegistry(object):
    extensions = []

    @staticmethod
    def add(extension, priority=0):
        ManifestExtensionRegistry.extensions.append((extension, priority))
        ManifestExtensionRegistry.extensions.sort(key=lambda x: x[1])

    @staticmethod
    def export_manifest(manifest, task):
        for extension, _ in ManifestExtensionRegistry.extensions:
            extension.export_manifest(manifest, task)

    @staticmethod
    def import_manifest(manifest):
        if not manifest.is_valid():
            return
        for extension, _ in ManifestExtensionRegistry.extensions:
            extension.import_manifest(manifest)

    @staticmethod
    def import_protobuf(pb):
        for extension, _ in ManifestExtensionRegistry.extensions:
            extension.import_protobuf(pb)


class ManifestExtension(object):
    def export_manifest(self, manifest, task):
        pass

    def import_manifest(self, manifest):
        pass

    def import_protobuf(self, pb):
        pass



import bz2
import copy
import getpass
import gzip
import lzma
import subprocess
import os
import platform
import sys
import threading
import time
if os.name != "nt":
    import termios
import glob
import multiprocessing
import re
import shutil
import tarfile
import zipfile
import bz2file
import hashlib
from contextlib import contextmanager
from psutil import NoSuchProcess, Process
from jinja2 import Environment, FileSystemLoader
from jinja2.exceptions import TemplateError
from jinja2.runtime import Context
from jinja2.utils import missing
from requests import Session
from requests.auth import HTTPBasicAuth
from urllib.parse import urlparse, urlunparse


from jolt import cache
from jolt import filesystem as fs
from jolt import log
from jolt import utils
from jolt import config
from jolt.error import JoltCommandError, JoltTimeoutError
from jolt.error import raise_error_if
from jolt.error import raise_task_error, raise_task_error_if


http_session = Session()


def stdout_write(line):
    sys.stdout.write(line + "\n")
    sys.stdout.flush()


def stderr_write(line):
    sys.stderr.write(line + "\n")
    sys.stderr.flush()


def _run(cmd, cwd, env, preexec_fn, *args, **kwargs):
    output = kwargs.get("output")
    output_on_error = kwargs.get("output_on_error")
    output_rstrip = kwargs.get("output_rstrip", True)
    output_stdio = kwargs.get("output_stdio", False)
    return_stderr = kwargs.get("return_stderr", False)
    output = output if output is not None else True
    output = False if output_on_error else output
    shell = kwargs.get("shell", True)
    timeout = kwargs.get("timeout", config.getint("jolt", "command_timeout", 0))
    timeout = timeout if type(timeout) is int and timeout > 0 else None

    log.debug("Running: '{0}' (CWD: {1})", cmd, cwd)

    p = subprocess.Popen(
        cmd,
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        shell=shell,
        cwd=cwd,
        env=env,
        preexec_fn=preexec_fn,
    )

    class Reader(threading.Thread):
        def __init__(self, parent, stream, output=None, logbuf=None):
            super(Reader, self).__init__()
            self.output = output
            self.parent = parent
            self.stream = stream
            self.logbuf = logbuf if logbuf is not None else []
            self.start()

        def run(self):
            line = ""
            try:
                with log.map_thread(self, self.parent):
                    for line in iter(self.stream.readline, b''):
                        if output_rstrip:
                            line = line.rstrip()
                        line = line.decode(errors='ignore')
                        if self.output:
                            self.output(line)
                        self.logbuf.append((self, line))
            except Exception as e:
                if self.output:
                    self.output("{0}", str(e))
                    self.output(line)
                self.logbuf.append((self, line))

    stdout_func = log.stdout if not output_stdio else stdout_write
    stderr_func = log.stderr if not output_stdio else stderr_write

    logbuf = []
    stdout = Reader(
        threading.current_thread(), p.stdout,
        output=stdout_func if output else None, logbuf=logbuf)
    stderr = Reader(
        threading.current_thread(), p.stderr,
        output=stderr_func if output else None, logbuf=logbuf)

    def terminate(pid):
        try:
            process = Process(pid)
            for chld in process.children(recursive=True):
                chld.terminate()
            process.terminate()
        except NoSuchProcess:
            pass

    def kill(pid):
        try:
            process = Process(pid)
            for chld in process.children(recursive=True):
                chld.kill()
            process.kill()
        except NoSuchProcess:
            pass

    timedout = False
    try:
        p.wait(timeout=timeout)
    except (subprocess.TimeoutExpired, JoltTimeoutError):
        timedout = True
        try:
            terminate(p.pid)
            p.wait(10)
        except subprocess.TimeoutExpired:
            kill(p.pid)
            p.wait()
    finally:
        stdout.join()
        stderr.join()
        p.stdin.close()
        p.stdout.close()
        p.stderr.close()

    if p.returncode != 0 and output_on_error:
        for reader, line in logbuf:
            if reader is stdout:
                log.stdout(line)
            else:
                log.stderr(line)

    stdoutbuf = []
    stderrbuf = []
    for reader, line in logbuf:
        if reader is stdout:
            stdoutbuf.append(line)
        else:
            stderrbuf.append(line)

    if p.returncode != 0:
        stderrbuf = [line for reader, line in logbuf if reader is stderr]
        if timedout:
            raise JoltTimeoutError(
                "Command timeout: {0}".format(
                    " ".join(cmd) if type(cmd) is list else cmd.format(*args, **kwargs)))
        else:
            raise JoltCommandError(
                "Command failed: {0}".format(
                    " ".join(cmd) if type(cmd) is list else cmd.format(*args, **kwargs)),
                stdoutbuf, stderrbuf, p.returncode)
    if return_stderr:
        return "\n".join(stdoutbuf) if output_rstrip else "".join(stdoutbuf), \
            "\n".join(stderrbuf) if output_rstrip else "".join(stderrbuf)
    return "\n".join(stdoutbuf) if output_rstrip else "".join(stdoutbuf)


class _String(object):
    def __init__(self, s=None):
        self._str = s or ''

    def __enter__(self):
        return self

    def __exit__(self, type, value, tb):
        pass

    def __str__(self):
        return self._str

    def __get__(self):
        return self._str

    def __add__(self, s):
        return self._str + s

    def __iadd__(self, s):
        self._str += s
        return self

    def endswith(self, substr):
        return self._str.endswith(substr)

    def startswith(self, substr):
        return self._str.startswith(substr)


class _CMake(object):
    def __init__(self, deps, tools, incremental=False):
        self.deps = deps
        self.tools = tools
        self.builddir = self.tools.builddir(incremental=incremental)
        self.installdir = self.tools.builddir("install", incremental=False)

    def configure(self, sourcedir, *args, generator=None, **kwargs):
        sourcedir = self.tools.expand_path(sourcedir)

        extra_args = list(args)
        extra_args += ["-D{0}={1}".format(key, self.tools.expand(val))
                       for key, val in kwargs.items()]
        extra_args = " ".join(extra_args)

        with self.tools.cwd(self.builddir):
            self.tools.run(
                "cmake {0} -B{1} -DCMAKE_INSTALL_PREFIX={2} {3} {4}",
                sourcedir,
                self.builddir,
                self.installdir,
                utils.option("-G", generator),
                extra_args,
                output=True)

    def build(self, release=True, *args, **kwargs):
        threading_args = ' -j {}'.format(kwargs.get("threads", self.tools.thread_count()))
        with self.tools.cwd(self.builddir):
            release = "--config Release" if release else ""
            self.tools.run("cmake --build . {0}{1}", release, threading_args, output=True)

    def install(self, release=True, *args, **kwargs):
        with self.tools.cwd(self.builddir):
            release = "--config Release" if release else ""
            self.tools.run("cmake --build . --target install {0}", release, output=True)

    def publish(self, artifact, files='*', *args, **kwargs):
        with self.tools.cwd(self.installdir):
            artifact.collect(files, *args, **kwargs)


class _Meson(object):
    def __init__(self, deps, tools):
        self.deps = deps
        self.tools = tools
        self.builddir = self.tools.builddir()
        self.installdir = self.tools.builddir("install")

    def configure(self, sourcedir, *args, **kwargs):
        sourcedir = self.tools.expand_path(sourcedir)
        self.tools.run("meson --prefix=/ {0} {1}", sourcedir, self.builddir,
                       output=True)

    def build(self, *args, **kwargs):
        self.tools.run("ninja -C {0} ", self.builddir, output=True)

    def install(self, *args, **kwargs):
        self.tools.run("DESTDIR={0} ninja -C {1} install",
                       self.installdir, self.builddir,
                       output=True)

    def publish(self, artifact, files='*', *args, **kwargs):
        with self.tools.cwd(self.installdir):
            artifact.collect(files, *args, **kwargs)


class _AutoTools(object):
    def __init__(self, deps, tools):
        self.deps = deps
        self.tools = tools
        self.builddir = self.tools.builddir()
        self.installdir = self.tools.builddir("install")

    def configure(self, sourcedir, *args, **kwargs):
        sourcedir = self.tools.expand_path(sourcedir)
        prefix = kwargs.get("prefix", "/")

        if not fs.path.exists(fs.path.join(sourcedir, "configure")):
            with self.tools.cwd(sourcedir):
                self.tools.run("autoreconf -visf", output=True)

        with self.tools.cwd(self.builddir), self.tools.environ(DESTDIR=self.installdir):
            self.tools.run("{0}/configure --prefix={1} {2}",
                           sourcedir, prefix,
                           self.tools.getenv("CONFIGURE_FLAGS", ""),
                           output=True)

    def build(self, *args, **kwargs):
        with self.tools.cwd(self.builddir), self.tools.environ(DESTDIR=self.installdir):
            self.tools.run("make VERBOSE=yes Q= V=1 -j{0}",
                           self.tools.cpu_count(), output=True)

    def install(self, target="install", **kwargs):
        with self.tools.cwd(self.builddir), self.tools.environ(DESTDIR=self.installdir):
            self.tools.run("make {}", target, output=True)

    def publish(self, artifact, files='*', *args, **kwargs):
        with self.tools.cwd(self.installdir):
            artifact.collect(files, *args, **kwargs)


class ZipFile(zipfile.ZipFile):
    """ ZipFile customization that preserves file permissions. """

    def extract(self, member, path=None, pwd=None):
        out_path = super().extract(member, path, pwd)

        # Restore permissions, if UNIX permissions are available
        info = self.getinfo(member)
        attr = info.external_attr >> 16
        if attr != 0:
            os.chmod(out_path, attr)

        return out_path

    def extractall(self, path=None, members=None, pwd=None):
        if members is None:
            members = self.namelist()

        for member in members:
            self.extract(member, path, pwd)


class _Tarfile(tarfile.TarFile):
    """ Tarfile customzation that can extract without uid/gids """

    def __init__(self, *args, ignore_owner=False, **kwargs):
        super().__init__(*args, **kwargs)
        self.__ignore_owner = ignore_owner

    def chown(self, *args, **kwargs):
        if self.__ignore_owner:
            return
        return super().chown(*args, **kwargs)


class JinjaTaskContext(Context):
    """
    Helper context for Jinja templates.

    Attempts to resolves any missing keywords by looking up task class attributes.
    """
    def resolve_or_missing(self, key):
        if key in self.vars:
            return self.vars[key]

        if key in self.parent:
            return self.parent[key]

        if key != "task":
            task = self.get("task")
            if task and hasattr(task, key):
                return getattr(task, key)

        return missing


class Namespace(object):
    def __init__(self, child=False):
        self.child = child

    def __enter__(self, *args):
        if not self.child:
            raise NamespaceException()
        return self

    def __exit__(self, type, exc, tb):
        if type == NamespaceException:
            return True
        return False


class NamespaceException(Exception):
    pass


def _subid(id, login):
    """ PIDs allowed to be mapped by a user. """
    with open(f"/etc/sub{id}") as subid:
        try:
            for line in subid:
                user, start, count = line.strip().split(":")
                if login == user:
                    return int(start), int(count)
        except Exception:
            pass
        return None, None


def _default_idmap(type, inner):
    """ Creates a default mapping of uid/gid between namespaces based
        on the requested inner uid/gid. As many ids as possible are mapped
        starting at 0. Requested ids are always mapped. """
    map = []
    outer = os.geteuid()
    start, count = _subid("uid", getpass.getuser())
    if start is None or count is None or count <= 0:
        return [(inner, outer, 1)]
    map.append((inner, outer, 1))
    if count <= 1:
        return map
    if inner == 0:
        map.append((inner + 1, start, count))
    else:
        map.append((0, start, min(inner - 1, count)))
        start += min(inner - 1, count) + 1
        count -= min(inner - 1, count) + 1
        if count > 0:
            map.append((inner + 1, start, count))
    return map


class Tools(object):
    """ A collection of useful tools.

    Any {keyword} arguments, or macros, found in strings passed to
    tool functions are automatically expanded to the value of the
    associated task's parameters and properties. Relative paths are
    made absolute by prepending the current working directory.
    """

    _builddir_lock = threading.RLock()

    def __init__(self, task=None, cwd=None, env=None):
        self._chroot = None
        self._chroot_prefix = []
        self._chroot_path = []
        self._deadline = None
        self._run_prefix = []
        self._preexec_fn = None
        self._cwd = fs.path.normpath(fs.path.join(config.get_workdir(), cwd or config.get_workdir()))
        self._env = copy.deepcopy(env or os.environ)
        self._task = task
        if task:
            self._env["JOLTDIR"] = task.joltdir
            self._env["JOLTBUILDDIR"] = self.buildroot
            self._env["JOLTCACHEDIR"] = config.get_cachedir()
        self._builddir = {}

    def __enter__(self):
        return self

    def __exit__(self, type, value, tb):
        for dir in self._builddir.values():
            fs.rmtree(dir, ignore_errors=True)
        return False

    def append_file(self, pathname, content, expand=True):
        """ Appends data at the end of a file.

        Args:
            pathname (str): Path to file. The file must exist.
            content (str): Data to be appended at the end of the file.
        """

        pathname = self.expand_path(pathname)
        if expand:
            content = self.expand(content)
        with open(pathname, "ab") as f:
            f.write(content.encode())

    def _make_zipfile(self, filename, fmt, rootdir):
        dirname = fs.path.dirname(filename)
        if not fs.path.exists(dirname):
            fs.makedirs(dirname)
        with zipfile.ZipFile(filename, "w", zipfile.ZIP_DEFLATED) as zf:
            for dirpath, dirnames, filenames in os.walk(rootdir):
                for name in sorted(dirnames):
                    path = os.path.normpath(os.path.join(dirpath, name))
                    zippath = os.path.relpath(path, rootdir)
                    zf.write(path, zippath)
                for name in filenames:
                    path = os.path.normpath(os.path.join(dirpath, name))
                    zippath = os.path.relpath(path, rootdir)
                    if os.path.isfile(path):
                        zf.write(path, zippath)
        return filename

    def _make_tarfile(self, filename, fmt, rootdir):
        dirname = os.path.dirname(filename)
        if not os.path.exists(dirname):
            fs.makedirs(dirname)
        with tarfile.open(filename, 'w|%s' % fmt) as tar:
            tar.add(rootdir, ".")
        return filename

    def archive(self, pathname, filename):
        """ Creates a (compressed) archive.

        The type of archive to create is determined by the filename extension.
        Supported formats are:

        - tar
        - tar.bz2
        - tar.gz
        - tar.xz
        - zip

        Args:
            pathname (str): Directory path of files to be archived.
            filename (str): Name/path of created archive.
        """
        filename = self.expand_path(filename)
        pathname = self.expand_path(pathname)

        fmt = None
        if filename.endswith(".zip"):
            fmt = "zip"
        elif filename.endswith(".tar"):
            fmt = "tar"
        elif filename.endswith(".tar.gz"):
            if self.which("tar") and self.which("pigz"):
                self.run("tar -I pigz -cf {} -C {} .", filename, pathname)
                return filename
            fmt = "targz"
        elif filename.endswith(".tgz"):
            if self.which("tar") and self.which("pigz"):
                self.run("tar -I pigz -cf {} -C {} .", filename, pathname)
                return filename
            fmt = "targz"
        elif filename.endswith(".tar.bz2"):
            fmt = "tarbz2"
        elif filename.endswith(".tar.xz"):
            fmt = "tarxz"
        raise_task_error_if(
            not fmt, self._task,
            "unknown archive type '{0}'", fs.path.basename(filename))
        try:
            if fmt == "zip":
                outfile = self._make_zipfile(filename, fmt, rootdir=pathname)
            else:
                outfile = self._make_tarfile(filename, fmt[3:], rootdir=pathname)
            if outfile != filename:
                shutil.move(outfile, filename)
            return filename
        except Exception:
            raise_task_error(self._task, "failed to create archive from directory '{0}'", pathname)

    def autotools(self, deps=None):
        """ Creates an AutoTools invokation helper """
        return _AutoTools(deps, self)

    @utils.locked(lock='_builddir_lock')
    def builddir(self, name=None, incremental=False, unique=True):
        """ Creates a temporary build directory.

        The build directory will persist for the duration of a task's
        execution. It is automatically removed afterwards.

        Args:
            name (str): Name prefix for the directory. A unique
                autogenerated suffix will also be appended to the
                final name.
            incremental (boolean): If false, the created directory is
                deleted upon completion of the task.

        Returns:
            Path to the created directory.
        """
        name = self.expand(name or "build")
        name = fs.path.join(self.buildroot, name)

        # Append task name
        if self._task is not None and unique:
            name += "-" + utils.canonical(self._task.short_qualified_name)

        dirname = fs.path.join(self.getcwd(), name)

        # Check if incremental build directories are disabled in the configuration
        if incremental not in ["always"] and not config.is_incremental_build():
            incremental = False

        if incremental:
            # Create a unique build directory for each task
            # and store the task name in a hidden file.
            if self._task is not None and unique:
                meta_task = fs.path.join(dirname, ".task")
                if not fs.path.exists(meta_task) \
                   or self.read_file(meta_task) != self._task.qualified_name:
                    fs.rmtree(dirname, ignore_errors=True)
                    fs.makedirs(dirname)

                # Remove the build directory if the task taint has changed (--force or --salt)
                if self._task.taint is not None:
                    meta = fs.path.join(dirname, ".taint")
                    if not fs.path.exists(meta) or self.read_file(meta) != str(self._task.taint):
                        fs.rmtree(dirname, ignore_errors=True)
                        fs.makedirs(dirname)
                        self.write_file(meta, str(self._task.taint))

                self.write_file(meta_task, self._task.qualified_name)
            else:
                fs.makedirs(dirname)
            return dirname

        if name not in self._builddir:
            fs.makedirs(dirname)
            self._builddir[name] = dirname

        return self._builddir[name]

    @property
    def buildroot(self):
        """ Return the root path of all build directories """
        from jolt.loader import JoltLoader
        return fs.path.normpath(JoltLoader.get().build_path)

    def checksum_file(self, filelist, concat=False, hashfn=hashlib.sha1, filterfn=None):
        """ Calculate a checksum of one or multiple files.

        Args:
            filelist (str,list): One or multiple files.
            concat (boolean): Concatenate files and return a single digest. If False,
                a list with one digest for each file is returned. Default: False.
            hashfn: The hash algorithm used. Any type which provides an update() and
                hexdigest() method is accepted. Default: hashlib.sha1
            filterfn: An optional data filter function. It is called repeatedly
                with each block of data read from files as its only argument.
                It should return the data to be included in the checksum.
                Default: None

        Returns:
            A list of checksum digests, or a single digest if files where concatenated.
        """
        files = [self.expand_path(fname) for fname in utils.as_list(filelist)]
        filterfn = filterfn or (lambda data: data)
        result = []
        checksum = hashfn()

        for fname in files:
            with open(fname, "rb") as f:
                for block in iter(lambda: f.read(0x10000), b''):
                    checksum.update(bytes(filterfn(block)))
                result.append(checksum.hexdigest())
            if not concat:
                checksum = hashfn()

        return result[-1] if concat or type(filelist) is str else result

    def chmod(self, pathname, mode):
        """ Changes permissions of files and directories.

        Args:
            pathname (str): Path to a file or directory to change
                permissions for.
            mode (int): Requested permission bits.
        """
        pathname = self.expand_path(pathname)
        return os.chmod(pathname, mode)

    def cmake(self, deps=None, incremental=False):
        """ Creates a CMake invokation helper """
        return _CMake(deps, self, incremental)

    def compress(self, src, dst):
        """ Compress a file.

        Supported formats are:

        - .bz2
        - .gz
        - .xz

        Args:
            src (str): Source file to be compressed.
            dst (str): Destination path for compressed file. The filename extension
                determines the compression algorithm used.
        """
        src = self.expand_path(src)
        dst = self.expand_path(dst)

        ext = dst.rsplit(".", 1)[-1]
        if ext == "bz2":
            with open(src, 'rb') as infp:
                with bz2.open(dst, 'wb') as outfp:
                    for block in iter(lambda: infp.read(0x10000), b''):
                        outfp.write(block)
        elif ext == "gz":
            if self.which("pigz"):
                return self.run("pigz -p {} {}", self.thread_count(), src)
            with open(src, 'rb') as infp:
                with gzip.open(dst, 'wb') as outfp:
                    for block in iter(lambda: infp.read(0x10000), b''):
                        outfp.write(block)
        elif ext == "xz":
            with open(src, 'rb') as infp:
                with lzma.open(dst, 'wb') as outfp:
                    for block in iter(lambda: infp.read(0x10000), b''):
                        outfp.write(block)

    def copy(self, src, dst, symlinks=False):
        """ Copies file and directories (recursively).

        The directory tree structure is retained when copying directories.

        Args:
            src (str): Path to a file or directory to be copied.
            dest (str): Destination path. If the string ends with a
                path separator a new directory will
                be created and source files/directories will be copied into
                the new directory. A destination without trailing path
                separator can be used to rename single files, one at a time.
            symlinks (boolean, optional): If True, symlinks are copied.
                The default is False, i.e. the symlink target is copied.
        """
        src = self.expand_path(src)
        dst = self.expand_path(dst)
        return fs.copy(src, dst, symlinks=symlinks)

    def cpu_count(self):
        """ The number of CPUs on the host.

        Returns:
            int: The number of CPUs on the host.
        """

        return multiprocessing.cpu_count()

    def thread_count(self):
        """ Number of threads to use for a task.

        Returns:
            int: number of threads to use.
        """
        threads = config.get("jolt", "threads", self.getenv("JOLT_THREADS", None))
        return int(threads) if threads else self.cpu_count()

    @contextmanager
    def cwd(self, pathname, *args):
        """ Change the current working directory to the specified path.

        This function doesn't change the working directory of the Jolt
        process. It only changes the working directory for tools within
        the tools object.

        Args:
            pathname (str): Path to change to.

        Example:

            .. code-block:: python

                with tools.cwd("subdir") as cwd:
                    print(cwd)
        """
        path = self.expand_path(fs.path.join(str(pathname), *args))
        prev = self._cwd
        try:
            raise_task_error_if(
                not fs.path.exists(path) or not fs.path.isdir(path),
                self._task,
                "failed to change directory to '{0}'", path)
            self._cwd = path
            yield fs.path.normpath(self._cwd)
        finally:
            self._cwd = prev

    def download(self, url, pathname, exceptions=True, auth=None, **kwargs):
        """
        Downloads a file using HTTP.

        Automatically expands any {keyword} arguments in the URL and pathname.

        Basic authentication is supported by including the credentials in the URL.
        Environment variables can be used to hide sensitive information. Specify
        the environment variable name in the URI as e.g.
        ``http://{environ[USER]}:{environ[PASS]}@host``.
        Alternatively, the auth parameter can be used to provide an authentication
        object that is passed to the requests.get() function.

        Throws a JoltError exception on failure.

        Args:
           url (str): URL to the file to be downloaded.
           pathname (str): Name/path of destination file.
           kwargs (optional): Addidional keyword arguments passed on
               directly ``requests.get()``.

        """

        url = self.expand(url)
        pathname = self.expand_path(pathname)

        url_parsed = urlparse(url)
        raise_task_error_if(
            not url_parsed.scheme or not url_parsed.netloc,
            self._task,
            "Invalid URL: '{}'", url)

        if auth is None and url_parsed.username and url_parsed.password:
            auth = HTTPBasicAuth(url_parsed.username, url_parsed.password)

        # Redact password from URL if present
        if url_parsed.password:
            url_parsed = url_parsed._replace(netloc=url_parsed.netloc.replace(url_parsed.password, "****"))

        url_cleaned = urlunparse(url_parsed)

        try:
            response = http_session.get(url, stream=True, auth=auth, **kwargs)
            raise_error_if(
                exceptions and response.status_code not in [200],
                f"Download from '{url_cleaned}' failed with status '{response.status_code}'")

            name = fs.path.basename(pathname)
            size = int(response.headers.get('content-length', 0))
            with log.progress("Downloading {0}".format(utils.shorten(name)), size, "B") as pbar:
                log.verbose("{} -> {}", url_cleaned, pathname)
                with open(pathname, 'wb') as out_file:
                    chunk_size = 4096
                    for data in response.iter_content(chunk_size=chunk_size):
                        out_file.write(data)
                        pbar.update(len(data))
                actual_size = self.file_size(pathname)
                raise_error_if(
                    size != 0 and size != actual_size,
                    f"Downloaded file was truncated to {actual_size}/{size} bytes: {name}")

            return response.status_code == 200
        except Exception as e:
            utils.call_and_catch(self.unlink, pathname)
            raise e

    @contextmanager
    def environ(self, **kwargs):
        """ Set/get environment variables.

        Only child processes spawned by the same tools object will be affected
        by the changed environment.

        The changed environment is only valid within a context and it is
        restored immediately upon leaving the context.

        Args:
            kwargs (optinal): A list of keyword arguments assigning values to
                environment variable.

        Example:

            .. code-block:: python

                with tools.environ(CC="clang"):
                    tools.run("make all")
        """
        restore = {key: value for key, value in self._env.items()}

        for key, value in kwargs.items():
            if value is not None:
                self._env[key] = self.expand(value)
            else:
                self._env.pop(key, None)

        try:
            yield self._env
        finally:
            self._env = restore

    def exists(self, pathname):
        """ Checks if a file or directory exists.

        Args:
            pathname (str): Path to file or directory.

        Returns:
            bool: True if the file or directory exists, False otherwise.
        """
        return fs.path.exists(self.expand_path(pathname))

    def expand(self, string, *args, **kwargs):
        """ Expands keyword arguments/macros in a format string.

        This function is identical to ``str.format()`` but it
        automatically collects keyword arguments from a task's parameters
        and properties.

        It also supports three additional conversion specifiers:

          - c: call method without arguments
          - l: convert string to lower case
          - u: convert string to upper case

        Args:
            string (str): The string to be expanded.
            args (str, optional): Additional positional values required
                by the format string.
            kwargs (str, optional): Additional keyword values required by
                the format string.

        Returns:
            str: Expanded string.

        Example:

            .. code-block:: python

                target = Parameter(default="all")
                verbose = "yes"

                def run(self, deps, tools):
                    print(tools.expand("make {target} VERBOSE={verbose}"))  # "make all VERBOSE=yes"

        """
        return self._task.expand(string, *args, **kwargs) \
            if self._task is not None \
            else utils.expand(string, *args, **kwargs)

    def expand_path(self, pathname, *args, **kwargs):
        """ Expands keyword arguments/macros in a pathname format string.

        This function is identical to ``str.format()`` but it
        automatically collects keyword arguments from a task's parameters
        and properties.

        The function also makes relative paths absolute by prepending the
        current working directory.

        Args:
            pathname (str): The pathname to be expanded.
            args (str, optional): Additional positional values required
                by the format pathname.
            kwargs (str, optional): Additional keyword values required by
                the format pathname.

        Return
            str: Expanded string.
        """

        if type(pathname) is list:
            return [self.expand_path(path) for path in pathname]

        path = fs.path.join(self.getcwd(), self.expand(pathname, *args, **kwargs))
        # Ensure to retain any trailing path separator which is used as
        # indicator of directory paths
        psep = fs.sep if path[-1] in fs.anysep else ""
        return fs.path.normpath(path) + psep

    def expand_relpath(self, pathname, relpath=None, *args, **kwargs):
        """ Expands keyword arguments/macros in a pathname format string.

        This function is identical to ``str.format()`` but it
        automatically collects keyword arguments from a task's parameters
        and properties.

        The function also makes absolute paths relative to a specified
        directory.

        Args:
            pathname (str): The pathname to be expanded.
            relpath (str, optional): Directory to which the returned path will be relative.
                If not provided, the ``joltdir`` attribute is used.
            args (str, optional): Additional positional values required
                by the format pathname.
            kwargs (str, optional): Additional keyword values required by
                the format pathname.

        Return
            str: Expanded string.
        """

        if not relpath:
            relpath = self._task.joltdir if self._task else self.getcwd()
        pathname = self.expand(pathname, *args, **kwargs)
        relpath = self.expand(relpath, *args, **kwargs)
        pathname = fs.path.join(self.getcwd(), pathname)
        # Ensure to retain any trailing path separator which is used as
        # indicator of directory paths
        psep = fs.sep if pathname[-1] in fs.anysep else ""
        return fs.path.relpath(pathname, relpath) + psep

    def extract(self, filename, pathname, files=None, ignore_owner=False):
        """ Extracts files in an archive.

        Supported formats are:

        - tar
        - tar.bz2
        - tar.gz
        - tar.xz
        - zip

        Args:
            filename (str): Name/path of archive file to be extracted.
            pathname (str): Destination path for extracted files.
            files (list, optional): List of files the be extracted
                from the archive. If not provided, all files are
                extracted.

        """
        filename = self.expand_path(filename)
        filepath = self.expand_path(pathname)
        ignore_owner_tar = "--no-same-owner" if ignore_owner else ""
        try:
            fs.makedirs(filepath)
            if filename.endswith(".zip"):
                with ZipFile(filename, 'r') as zip:
                    zip.extractall(filepath, files)
            elif filename.endswith(".tar"):
                with _Tarfile.open(filename, 'r', ignore_owner=ignore_owner) as tar:
                    if files:
                        for file in files:
                            tar.extract(file, filepath)
                    else:
                        tar.extractall(filepath)
            elif filename.endswith(".tar.gz") or filename.endswith(".tgz"):
                if self.which("tar") and self.which("pigz"):
                    self.run("tar -I pigz {} -xf {} -C {} {}",
                             ignore_owner_tar, filename, filepath,
                             " ".join(files) if files else "")
                    return
                with _Tarfile.open(filename, 'r:gz', ignore_owner=ignore_owner) as tar:
                    if files:
                        for file in files:
                            tar.extract(file, filepath)
                    else:
                        tar.extractall(filepath)
            elif filename.endswith(".tar.bz2"):
                # bz2file module for multistream support
                with bz2file.open(filename) as bz2:
                    with _Tarfile.open(fileobj=bz2, ignore_owner=ignore_owner) as tar:
                        if files:
                            for file in files:
                                tar.extract(file, filepath)
                        else:
                            tar.extractall(filepath)
            elif filename.endswith(".tar.xz"):
                with _Tarfile.open(filename, 'r:xz', ignore_owner=ignore_owner) as tar:
                    if files:
                        for file in files:
                            tar.extract(file, filepath)
                    else:
                        tar.extractall(filepath)
            else:
                raise_task_error(self._task, "unknown archive type '{0}'", fs.path.basename(filename))
        except Exception:
            log.exception()
            raise_task_error(self._task, "failed to extract archive '{0}'", filename)

    def file_size(self, pathname):
        """ Determines the size of a file.

        Args:
            pathname (str): Name/path of file for which the size is requested.

        Returns:
            int: The size of the file in bytes.
        """
        pathname = self.expand_path(pathname)
        try:
            stat = os.stat(pathname)
        except KeyboardInterrupt as e:
            raise e
        except Exception:
            raise_task_error(self._task, "file not found '{0}'", pathname)
        else:
            return stat.st_size

    def getcwd(self):
        """ Returns the current working directory. """
        return fs.path.normpath(self._cwd)

    def getenv(self, key, default=None):
        """ Returns the value of an environment variable.

        Only child processes spawned by the same tools object can see
        the environment variables and their values returned by this method.
        Don't assume the same variables are set in the Jolt process' environment.
        """
        return self._env.get(key, default)

    def glob(self, pathname, expand=False):
        """ Enumerates files and directories.

        Args:
            pathname (str): A pathname pattern used to match files to be
                included in the returned list of files and directories.
                The pattern may contain simple shell-style
                wildcards such as '*' and '?'. Note: files starting with a
                dot are not matched by these wildcards.
            expand (boolean): Expand matches to absolute paths. Default: false.

        Returns:
            A list of file and directory pathnames. The pathnames are relative
            to the current working directory unless the ``pathname`` argument
            was absolute.

        Example:

            .. code-block:: python

                textfiles = tools.glob("*.txt")
        """
        path = self.expand_path(pathname)
        files = utils.as_list(glob.glob(path, recursive=True))
        if expand:
            files = [self.expand_path(file) for file in files]
        elif not fs.path.isabs(pathname):
            files = [self.expand_relpath(file, self.getcwd()) for file in files]
        return list(sorted(files))

    def isdir(self, pathname):
        """ Determines if a path is a directory.

        Args:
            pathname (str): Path to a file or directory.

        Returns:
            boolean: True if the path is a directory, False otherwise.
        """
        pathname = self.expand_path(pathname)
        return fs.path.isdir(pathname)

    def mkdir(self, pathname, recursively=True):
        """ Create directory. """

        pathname = self.expand_path(pathname)
        if recursively:
            fs.makedirs(pathname)
        else:
            fs.mkdir(pathname)

    def mkdirname(self, pathname, recursively=True):
        """ Create parent directory. """

        pathname = self.expand_path(pathname)
        pathname = fs.path.dirname(pathname)
        if pathname:
            self.mkdir(pathname, recursively)

    def move(self, src, dst):
        """
        Move/rename file.

        Args:
            src (str): Path to a file or directory to be moved.
            dest (str): Destination path. If the destination is
                an existing directory, then src is moved inside
                that directory. If the destination already exists
                but is not a directory, it may be overwritten.
                If the destination is not on the same filesystem, the
                source file or directory is copied to the destination
                and then removed.
        """

        src = self.expand_path(src)
        dst = self.expand_path(dst)
        return shutil.move(src, dst)

    def map_consecutive(self, callable, iterable):
        """ Same as ``map()``. """
        return utils.map_consecutive(callable, iterable)

    def map_concurrent(self, callable, iterable, max_workers=None):
        """ Concurrent `~map()`.

        Args:
            callable: A callable object to be executed for each item in
                the collection.
            iterable: An iterable collection of items.
            max_workers (optional): The maximum number of worker threads
                allowed to be spawned when performing the work. The
                default is the value returned by
                :func:`jolt.Tools.cpu_count()`.

        Returns:
            list: List of return values.


        Example:

            .. code-block:: python

                def compile(self, srcfile):
                    objfile = srcfile + ".o"
                    tools.run("gcc -c {0} -o {1}", srcfile, objfile)
                    return objfile

                srcfiles = ["test.c", "main.c"]
                objfiles = tools.map_concurrent(compile, srcfiles)

        """
        return utils.map_concurrent(callable, iterable, max_workers)

    def meson(self, deps=None):
        """ Creates a Meson invokation helper """
        return _Meson(deps, self)

    @contextmanager
    def nixpkgs(self, nixfile=None, packages=None, pure=False, path=None, options=None):
        """
        Creates a Nix environment with the specified packages.

        Args:
            nixfile (str): Path to a Nix expression file.
            packages (list): List of Nix packages to include in environment.
            pure (boolean): Create a pure environment.
            path (list): List of Nix expression paths.
            options (dict): Nix configuration options.

        Example:

            .. code-block:: python

                def run(self, deps, tools):
                    with tools.nixpkgs(packages=["gcc13"]):
                        tools.run("gcc --version")

        """

        # Check if Nix is available
        raise_task_error_if(
            not self.which("nix-shell"),
            self._task,
            "Nix not available on this system")

        nixfile = self.expand_path(nixfile) if nixfile else ""
        pathflags = " ".join([f"-I {path}" for path in path or []])
        options = " ".join([f"--option {k} {v}" for k, v in (options or {}).items()])
        pureflag = "--pure" if pure else ""
        packages = "-p " + " ".join(packages) if packages else ""

        # Expand all placeholders
        options = self.expand(options)
        packages = self.expand(packages)
        pathflags = self.expand(pathflags)

        # Run nix-shell to stage packages and environment
        env = self.run(
            "nix-shell {} {} {} --run 'env -0' {}",
            pathflags,
            pureflag,
            packages,
            nixfile,
            output_on_error=True)
        env = env.strip().strip("\x00")
        env = dict(line.split("=", 1) for line in env.split('\x00'))

        # Add host path first to environment PATH
        host_path = env.get("HOST_PATH", None)
        if host_path:
            env["PATH"] = host_path + os.pathsep + env["PATH"]

        # Enter the environment
        old_env = self._env
        try:
            if pure:
                self._env = env
            else:
                self._env = copy.deepcopy(env)
                self._env.update(env)
            yield
        finally:
            self._env = old_env

    def render(self, template, **kwargs):
        """ Render a Jinja template string.

        Args:
            template (str): Jinja template string.
            kwargs (dict): Keywords made available to the template context.
               Task attributes are automatically available.

        Returns:
            str: Renderered template data.

        """
        try:
            env = Environment(
                loader=FileSystemLoader(self.getcwd()),
                autoescape=False,
                trim_blocks=True,
                lstrip_blocks=True)
            env.context_class = JinjaTaskContext
            env.filters["prefix"] = utils.prefix
            env.filters["suffix"] = utils.suffix
            tmpl = env.from_string(template)
            return tmpl.render(task=self._task, tools=self, **kwargs)
        except TemplateError as e:
            log.debug("Template error: {}", template)
            raise_task_error(self._task, "Template error: {}", e)

    def render_file(self, template, **kwargs):
        """ Render a Jinja template file.

        Args:
            template (str): Filesystem path to template file.
            kwargs (dict): Keywords made available to the template context.
               Task attributes are automatically available.

        Returns:
            str: Renderered template data.

        """
        try:
            env = Environment(
                loader=FileSystemLoader(self.getcwd()),
                autoescape=False,
                trim_blocks=True,
                lstrip_blocks=True)
            env.context_class = JinjaTaskContext
            tmpl = env.get_template(self.expand(template))
            return tmpl.render(task=self._task, tools=self, **kwargs)
        except TemplateError as e:
            log.debug("Template error: {}", template)
            raise_task_error(self._task, "Template error: {}", e)

    def replace_in_file(self, pathname, search, replace, regex=False):
        """ Replaces all occurrences of a substring in a file.

        Args:
            pathname (str): Name/path of file to modify.
            search (str): Substring to be replaced.
            replace (str): Replacement substring.
            regex (boolean): Interpret search parameter as
                a regular expression matching the string to
                be replaced.

        Example:

            .. code-block:: python

                version = Parameter(default="1.0")

                def run(self, deps, tools):
                    tools.replace_in_file("Makefile", "VERSION := 0.9", "VERSION := {version}")
        """
        pathname = self.expand_path(pathname)
        search = self.expand(search)
        replace = self.expand(replace)
        try:
            with open(pathname, "rb") as f:
                data = f.read()
            if regex:
                data = re.sub(search.encode(), replace.encode(), data)
            else:
                data = data.replace(search.encode(), replace.encode())
            with open(pathname, "wb") as f:
                f.write(data)
        except KeyboardInterrupt as e:
            raise e
        except Exception:
            raise_task_error(self._task, "Failed to replace string in file '{0}'", pathname)

    def rmtree(self, pathname, *args, **kwargs):
        """Removes a directory tree from disk.

        Args:
            pathname (str): Path to the file or directory to be removed.
            ignore_errors (boolean, optional): Ignore files that can't be deleted.
                The default is ``False``.

        """
        pathname = self.expand_path(pathname, *args, **kwargs)
        return fs.rmtree(pathname, **kwargs)

    def rsync(self, srcpath, dstpath, *args, **kwargs):
        """ Synchronizes files from one directory to another.

        The function performs a smart copy of files from the
        ``srcpath`` directory to the ``dstpath`` directory in
        such a way that ``dstpath`` will mirror ``srcpath``.

        If ``dstpath`` is empty, the files are copied normally.

        If ``dstpath`` already contains a sub or superset of the
        files in ``srcpath``, files are either copied or deleted
        depending on their presence in the source directory. Common
        files are only copied if the file content differs, thereby
        retaining metadata (such as timestamps) of identical files
        already present in ``dstpath``.

        Args:
            srcpath (str): Path to source directory.
                The directory must exist.
            dstpath (str): Path to destination directory.

        """

        def _scandir(scanpath, filterfn=lambda path: path[0] != "."):
            def relresult(path, fp):
                return os.path.relpath(os.path.join(path, fp), scanpath)
            resfn = relresult
            result = []

            for path, dirs, files in os.walk(scanpath):
                for d in dirs:
                    if filterfn(d):
                        result.append((True, resfn(path, d)))
                for f in files:
                    if filterfn(f):
                        result.append((False, resfn(path, f)))

            return result

        srcpath = self.expand_path(srcpath, *args, **kwargs)
        dstpath = self.expand_path(dstpath, *args, **kwargs)
        srcfiles = set(_scandir(srcpath))
        dstfiles = set(_scandir(dstpath))
        added_files = list(srcfiles - dstfiles)
        deleted_files = list(dstfiles - srcfiles)
        common_files = srcfiles.intersection(dstfiles)

        # Remove files first, then directories
        for dir, fp in sorted(deleted_files, key=lambda n: (n[0], -len(n[1]))):
            dst = fs.path.join(dstpath, fp)
            if dir:
                fs.rmtree(dst)
            else:
                fs.unlink(dst)

        # Add new directories, then files
        for dir, fp in sorted(added_files, key=lambda n: (not n[0], n[1])):
            src = fs.path.join(srcpath, fp)
            dst = fs.path.join(dstpath, fp)
            if dir:
                fs.makedirs(dst)
            else:
                fs.copy(src, dst, metadata=False)

        # Refresh existing files
        for dir, fp in filter(lambda n: not n[0], common_files):
            src = fs.path.join(srcpath, fp)
            dst = fs.path.join(dstpath, fp)
            if not fs.identical_files(src, dst):
                fs.copy(src, dst, metadata=False)

    def run(self, cmd, *args, **kwargs):
        """
        Runs a command in a shell interpreter.

        These additional environment variables will be set when the command is run:

            - ``JOLTDIR`` - Set to :attr:`Task.joltdir <jolt.tasks.Task.joltdir>`
            - ``JOLTCACHEDIR`` - Set to the location of the Jolt cache

        A JoltCommandError exception is raised on failure.

        Args:
            cmd (str): Command format string.
            args: Positional arguments for the command format string.
            kwargs: Keyword arguments for the command format string.
            output (boolean, optional): By default, the executed command's
                output will be written to the console. Set to ``False`` to
                disable all output.
            output_on_error (boolean, optional): If ``True``, no output is
                written to the console unless the command fails.
                The default is ``False``.
            output_rstrip (boolean, optional): By default, output written
                to stdout is stripped from whitespace at the end of the
                string. This can be disabled by setting this argument to False.
            shell (boolean, optional): Use a shell to run the command.
                Default: True.
            timeout (int, optional): Timeout in seconds. The command will
                first be terminated if the timeout expires. If the command
                refuses to terminate, it will be killed after an additional
                10 seconds have passed. Default: None.

        Example:

            .. code-block:: python

                target = Parameter(default="all")
                verbose = "yes"

                def run(self, deps, tools):
                    tools.run("make {target} VERBOSE={verbose} JOBS={0}", tools.cpu_count())

        """
        kwargs.setdefault("shell", True)

        # Append command prefix before expanding string
        if self._chroot_prefix or self._run_prefix:
            if type(cmd) is list:
                cmd = self._chroot_prefix + self._run_prefix + cmd
            else:
                cmd = " ".join(self._chroot_prefix + self._run_prefix) + " " + cmd

        if self._deadline is not None:
            remaining = int(self._deadline - time.time() + 0.5)
            timeout = kwargs.get("timeout", remaining)
            kwargs["timeout"] = min(remaining, timeout)

        cmd = self.expand(cmd, *args, **kwargs)

        stdi, stdo, stde = None, None, None
        try:
            stdi, stdo, stde = None, None, None
            try:
                stdi = termios.tcgetattr(sys.stdin.fileno())
                stdo = termios.tcgetattr(sys.stdout.fileno())
                stde = termios.tcgetattr(sys.stderr.fileno())
            except KeyboardInterrupt as e:
                raise e
            except Exception:
                pass

            return _run(cmd, self._cwd, self._env, self._preexec_fn, *args, **kwargs)

        finally:
            if stdi:
                termios.tcsetattr(sys.stdin.fileno(), termios.TCSANOW, stdi)
            if stdo:
                termios.tcsetattr(sys.stdout.fileno(), termios.TCSANOW, stdo)
            if stde:
                termios.tcsetattr(sys.stderr.fileno(), termios.TCSANOW, stde)

    @contextmanager
    def runprefix(self, cmdprefix, *args, **kwargs):
        """
        Adds a command prefix to all commands executed by :func:`~run`.

        A new prefix is appended to any existing prefix.

        Args:
            cmdprefix (str, list): The command prefix. The string, or list,
                is expanded with :func:`~expand`.
            args (str, optional): Additional positional values passed to
                :func:`~expand`.
            kwargs (str, optional): Additional keyword values passed to
                :func:`~expand`.

        Example:

          .. code-block:: python

            with tools.runprefix("docker exec container"):
                tools.run("ls")


        The above code is equivalent to:

          .. code-block:: python

            tools.run("docker exec container ls")

        """
        cmdprefix = self.expand(cmdprefix, *args, **kwargs)
        if type(cmdprefix) is str:
            cmdprefix = cmdprefix.split()

        old_prefix = copy.copy(self._run_prefix)
        self._run_prefix += cmdprefix
        try:
            yield
        finally:
            self._run_prefix = old_prefix

    @utils.locked(lock='_builddir_lock')
    def sandbox(self, artifact, incremental=False, reflect=False):
        """ Creates a temporary build directory populated with the contents of an artifact.

        Files are copied using rsync.

        Args:
            artifact (cache.Artifact): A task artifact to be copied
                into the sandbox.
            incremental (boolean): If false, the created directory is
                deleted upon completion of the task.
            reflect (boolean): If true, a virtual sandbox is constructed
                from artifact metadata only. Files are not copied, but
                instead symlinks are created pointing at the origin of each
                file contained within the artifact. The sandbox reflects
                the artifact with a live view of the the current workspace.

        Returns:
            str: Path to the build directory..

        Example:

            .. code-block:: python

              def run(self, deps, tools):
                 sandbox = tools.sandbox(deps["boost"], incremental=True)
        """

        raise_error_if(
            type(artifact) is not cache.Artifact,
            "non-artifact passed as argument to Tools.sandbox()")

        suffix = utils.canonical(artifact.task.short_qualified_name)

        if reflect:
            sandbox_name = "sandboxes-reflected/" + suffix
        else:
            sandbox_name = "sandboxes/" + suffix

        path = self.builddir(sandbox_name, incremental=incremental, unique=False)
        if reflect:
            return self._sandbox_reflect(artifact, path)
        return self._sandbox_rsync(artifact, path)

    def _sandbox_validate(self, artifact, path):
        meta = fs.path.join(self.getcwd(), path, ".artifact")
        return meta if not fs.path.exists(meta) or self.read_file(meta) != artifact.path else None

    def _sandbox_rsync(self, artifact, path):
        meta = self._sandbox_validate(artifact, path)
        if meta:
            fs.unlink(meta, ignore_errors=True)
            self.rsync(artifact.path, path)
            self.write_file(meta, artifact.path)
        return path

    def _sandbox_reflect(self, artifact, path):
        meta = self._sandbox_validate(artifact, path)
        if meta:
            fs.rmtree(path)
            fs.makedirs(path)
            for relsrcpath, reldstpath in artifact.files.items():
                srcpath = fs.path.normpath(fs.path.join(artifact.task.joltdir, relsrcpath))
                dstpath = fs.path.normpath(fs.path.join(path, reldstpath))
                if dstpath != fs.path.realpath(dstpath):
                    log.debug("Cannot symlink '{} -> {}', parent directory already symlinked",
                              srcpath, dstpath)
                    continue
                if fs.path.isdir(dstpath):
                    files = fs.scandir(srcpath)
                    for file in files:
                        relfile = file[len(srcpath) + 1:]
                        self.symlink(file, fs.path.join(dstpath, relfile))
                else:
                    self.symlink(srcpath, dstpath)

                # Restore missing srcfiles if they resided in a build directory
                if srcpath.startswith(artifact.tools.buildroot) and \
                   not fs.path.exists(srcpath):
                    fs.copy(fs.path.join(artifact.path, reldstpath), srcpath, symlinks=True)
            self.write_file(meta, artifact.path)
        return path

    def setenv(self, key, value=None):
        """ Sets or unset an environment variable.

        Only child processes spawned by the same tools object can see
        the set environment variable and its value.
        Don't assume the same variable is set in the Jolt process' environment.

        Args:
            key (str): Name of variable to set.
            value (str): Value of the variable or ``None`` to unset it.

        """
        if value is None:
            try:
                del self._env[key]
            except KeyboardInterrupt as e:
                raise e
            except Exception:
                pass
        else:
            self._env[key] = self.expand(value)

    def symlink(self, src, dst, replace=True, relative=True):
        """ Creates a symbolic link.

        Args:
            src (str): Path to target file or directory.
            dst (str): Name/path of symbolic link.
            replace (boolean): Replace existing file or link. Default: false.
            relative (boolean): Create link using relative path to target.
                Default: false (absolute path).
        """
        src = self.expand_path(src) if not relative else self.expand(src)
        dst = self.expand_path(dst)
        dstdir = fs.path.dirname(dst) if dst[-1] != fs.sep else dst
        if replace and fs.path.lexists(dst):
            self.unlink(dst)
        if not fs.path.isdir(dstdir):
            fs.makedirs(dstdir)
        fs.symlink(src, dst)

    @contextmanager
    def timeout(self, seconds):
        """ Context manager to set a timeout for a block of code.

        A TimeoutError exception is raised if the block of code does not
        complete within the specified time.

        Args:
            seconds (int): Timeout in seconds.

        Example:

                .. code-block:: python

                    with tools.timeout(5):
                        tools.run("sleep 10")

        """
        if seconds is None:
            yield
            return

        with utils.timeout(seconds, JoltTimeoutError):
            old_deadline = self._deadline
            try:
                if old_deadline is None:
                    self._deadline = time.time() + seconds
                else:
                    self._deadline = min(old_deadline, time.time() + seconds)
                yield
            finally:
                self._deadline = old_deadline

    @contextmanager
    def tmpdir(self, name=None):
        """ Creates a temporary directory.

        The directory is only valid within a context and it is removed
        immediately upon leaving the context.

        Args:
            name (str): Name prefix for the directory. A unique
                autogenerated suffix will also be appended to the
                final name.

        Example:

            .. code-block:: python

                with tools.tmpdir() as tmp, tools.cwd(tmp):
                    tools.write_file("tempfile", "tempdata")

        """
        dirname = None
        try:
            self.mkdir(self.buildroot)
            dirname = fs.mkdtemp(prefix=(name or "tmpdir") + "-", dir=self.buildroot)
            yield fs.path.normpath(dirname)
        finally:
            if dirname:
                self.rmtree(dirname, ignore_errors=True)

    def unlink(self, pathname, *args, **kwargs):
        """Removes a file from disk.

        To remove directories, use :func:`~jolt.Tools.rmtree`.

        Args:
            pathname (str): Path to the file to be removed.

        """
        pathname = self.expand_path(pathname, *args, **kwargs)
        return fs.unlink(pathname, ignore_errors=kwargs.get("ignore_errors", False))

    @contextmanager
    @utils.deprecated
    def chroot(self, chroot, *args, path=None, **kwargs):
        """
        Experimental: Use chroot as root filesystem when running commands.

        Mounts the specified chroot as the root filesystem in a new Linux namespace,
        which is used when calling Tools.run().

        Requires a Linux host.

        Args:
            chroot (str, artifact): Path to rootfs directory, or an artifact
                with a 'rootfs' metadata path (artifact.paths.rootfs).
            path (list): List of directory paths within the chroot to add to
                the PATH environment variable, e.g. ["/usr/bin", "/bin"].
                By default, the current PATH is used also within the chroot.

        Example:

            .. code-block:: python

              with tools.choot("path/to/rootfs"):
                  tools.run("ls")

        """
        raise_error_if(platform.system() != "Linux", "Tools.chroot() is only supported on Linux")

        raise_task_error_if(
            not self.which("newuidmap") and not self.which("/usr/bin/newuidmap"), self._task,
            "No usable 'newuidmap' found in PATH")

        raise_task_error_if(
            not self.which("newgidmap") and not self.which("/usr/bin/newuidmap"), self._task,
            "No usable 'newgidmap' found in PATH")

        if type(chroot) is cache.Artifact:
            raise_task_error_if(
                not str(chroot.paths.rootfs), self._task,
                "No 'rootfs' path in artifact")
            chroot = chroot.paths.rootfs

        chroot = self.expand_path(chroot, *args, **kwargs)
        raise_task_error_if(
            not fs.path.exists(chroot) or not fs.path.isdir(chroot),
            self._task, "failed to change root to '{0}'", chroot)

        bind = []

        mount_dev = kwargs.get("mount_dev", True)
        mount_etc = kwargs.get("mount_etc", True)
        mount_home = kwargs.get("mount_home", False)
        mount_proc = kwargs.get("mount_proc", True)
        mount_joltdir = kwargs.get("mount_joltdir", True)
        mount_cachedir = kwargs.get("mount_cachedir", True)
        mount_builddir = kwargs.get("mount_builddir", True)
        mount = kwargs.get("mount", [])
        raise_task_error_if(
            type(mount) is not list,
            self._task, "Expected a list as mount argument to Tools.chroot()")
        mount = [self.expand(m) for m in mount]

        if mount_etc:
            bind.append("/etc/group")
            bind.append("/etc/hostname")
            bind.append("/etc/hosts")
            bind.append("/etc/passwd")
            bind.append("/etc/resolv.conf")

        if mount_home:
            bind.append("/home")

        if mount_joltdir and self._task:
            from jolt.loader import get_workspacedir
            bind.append(get_workspacedir())

        if mount_cachedir:
            bind.append(config.get_cachedir())

        if mount_builddir:
            bind.append(self.buildroot)

        if mount:
            for m in mount:
                bind.append(m)

        if mount_dev:
            bind.append("/dev")

        if mount_proc:
            bind.append("/proc")

        unshare = os.path.join(os.path.dirname(__file__), "chroot.py")

        old_chroot = self._chroot
        old_chroot_path = self._chroot_path
        old_chroot_prefix = self._chroot_prefix
        self._chroot = chroot

        if path:
            self._chroot_path = path
        else:
            self._chroot_path = self._env.get("PATH")
            self._chroot_path = self._chroot_path.split(fs.pathsep) if self._chroot_path else []

        try:
            with self.tmpdir("chroot") as bindroot:
                self._chroot_prefix = [
                    sys.executable,
                    unshare,
                    "-b",
                ] + bind + [
                    "-c",
                    chroot,
                    "-t",
                    bindroot,
                    "--shell={shell}",
                    "--",
                ]
                yield
        finally:
            self._chroot = old_chroot
            self._chroot_path = old_chroot_path
            self._chroot_prefix = old_chroot_prefix

    def _unshare(self, uidmap, gidmap):
        from ctypes import CDLL
        libc = CDLL("libc.so.6")

        CLONE_NEWNS = 0x00020000
        CLONE_NEWUSER = 0x10000000

        uidmap = [str(id) for map in uidmap for id in map]
        gidmap = [str(id) for map in gidmap for id in map]
        newuidmap = self.which("newuidmap")
        raise_task_error_if(
            not newuidmap, self._task,
            "No usable 'newuidmap' found in PATH")

        newgidmap = self.which("newgidmap")
        raise_task_error_if(
            not newgidmap, self._task,
            "No usable 'newgidmap' found in PATH")

        sem = multiprocessing.Semaphore(0)
        parent = os.getpid()
        child = os.fork()
        if child == 0:
            sem.acquire()
            pid = os.fork()
            if pid == 0:
                os.execve(newuidmap, ["newuidmap", str(parent)] + uidmap, {})
                os._exit(1)
            os.waitpid(pid, 0)
            os.execve(newgidmap, ["newgidmap", str(parent)] + gidmap, {})
            os._exit(1)
        assert libc.unshare(CLONE_NEWNS | CLONE_NEWUSER) == 0
        sem.release()
        os.waitpid(child, 0)

    @contextmanager
    def unshare(self, uid=0, gid=0, groups=[0], uidmap=None, gidmap=None):
        """
        Experimental: Create a Linux namespace.

        This method yields a new Linux namespace in which Python code may be executed.
        By default, the current user is mapped to root inside the namespace and all
        other users and groups are automatically mapped to the other user's configured
        subuids and subgids.

        The main use-case for namespaces is to fake the root user which may be useful
        in a number of situations:

          - to allow chroot() without beeing root
          - to allow mount() without beeing root
          - to preserve file ownership after tar file extraction
          - etc...

        Requires a Linux host.

        Note that the fork() system call is used. Changes made to variables
        will not persist after leaving the namespace.

        Args:
           uid (int): Requested uid inside the namespace. This is always mapped
               to the uid of the caller.
           gid (int): Requested gid inside the namespace. This is always mapped
               to the gid of the caller.
           uidmap (list): List of uids to map in the namespace. A list of tuples
               is expected: (inner uid, outer id, number of ids to map).
           gidmap (list): List of gids to map in the namespace. A list of tuples
               is expected: (inner gid, outer id, number of ids to map).

        Example:

            .. code-block:: python

              with tools.unshare() as ns, ns:
                  # Inside namespace
                  tools.run("whoami")  # "root"
              # Back outside namespace, namespace destructed

        """

        gidmap = gidmap or _default_idmap("gid", gid)
        uidmap = uidmap or _default_idmap("uid", uid)

        raise_task_error_if(
            not gidmap, self._task,
            "Invalid gid map: {}", gidmap)
        raise_task_error_if(
            not uidmap, self._task,
            "Invalid uid map: {}", uidmap)
        raise_task_error_if(
            not self.which("newuidmap"), self._task,
            "No usable 'newuidmap' found in PATH")
        raise_task_error_if(
            not self.which("newgidmap"), self._task,
            "No usable 'newgidmap' found in PATH")

        msgq = multiprocessing.JoinableQueue()
        pid = os.fork()
        if pid == 0:
            try:
                self._unshare(uidmap, gidmap)
                os.setuid(uid)
                os.setgid(gid)
                # if uid == 0 and gid == 0:
                #     os.setgroups(groups)
                yield Namespace(msgq)
            except Exception as exc:
                msgq.put(exc)
            else:
                msgq.put(None)
            msgq.join()
            os._exit(0)
        try:
            yield Namespace()
        except NamespaceException:
            pass
        exc = msgq.get()
        msgq.task_done()
        os.waitpid(pid, 0)
        if exc:
            raise exc

    def upload(self, pathname, url, exceptions=True, auth=None, **kwargs):
        """
        Uploads a file using HTTP (PUT).

        Automatically expands any {keyword} arguments in the URL and pathname.

        Basic authentication is supported by including the credentials in the URL.
        Environment variables can be used to hide sensitive information. Specify
        the environment variable name in the URI as e.g.
        ``http://{environ[USER]}:{environ[PASS]}@host``.
        Alternatively, the auth parameter can be used to provide an authentication
        object that is passed to the requests.get() function.

        Throws a JoltError exception on failure.

        Args:
           pathname (str): Name/path of file to be uploaded.
           url (str): Destination URL.
           auth (requests.auth.AuthBase, optional): Authentication helper.
               See requests.auth for details.
           kwargs (optional): Addidional keyword arguments passed on
               directly to `~requests.put()`.

        """
        pathname = self.expand_path(pathname)
        url = self.expand(url)
        name = fs.path.basename(pathname)
        size = self.file_size(pathname)

        url_parsed = urlparse(url)
        raise_task_error_if(
            not url_parsed.scheme or not url_parsed.netloc,
            self._task,
            "Invalid URL: '{}'", url)

        if auth is None and url_parsed.username and url_parsed.password:
            auth = HTTPBasicAuth(url_parsed.username, url_parsed.password)

        # Redact password from URL if present
        if url_parsed.password:
            url_parsed = url_parsed._replace(netloc=url_parsed.netloc.replace(url_parsed.password, "****"))

        url_cleaned = urlunparse(url_parsed)

        with log.progress("Uploading " + utils.shorten(name), size, "B") as pbar, \
             open(pathname, 'rb') as fileobj:
            log.verbose("{} -> {}", pathname, url_cleaned)

            def read():
                data = fileobj.read(4096)
                pbar.update(len(data))
                return data

            response = http_session.put(url, data=iter(read, b''), auth=auth, **kwargs)
            raise_error_if(
                exceptions and response.status_code not in [201, 204],
                f"Upload to '{url_cleaned}' failed with status '{response.status_code}'")
        return response.status_code in [201, 204]

    def read_file(self, pathname, binary=False):
        """ Reads a file. """
        pathname = self.expand_path(pathname)
        with open(pathname, "rb" if binary else "r") as f:
            return f.read()

    def read_depfile(self, pathname):
        """
        Reads a Make dependency file.

        Returns:
            dict: Dictionary of files and their dependencies.
        """
        pathname = self.expand_path(pathname)
        with open(pathname) as f:
            data = f.read()

        data = data.strip()
        data = data.replace("\\\n", "")
        data = data.splitlines()

        deps = {}

        for line in data:
            # Skip empty lines and comments
            if not line or line[0] == "#":
                continue

            parts = line.split(":", 1)
            raise_error_if(len(parts) != 2, "Depfile parse error: '{}'", line)
            outputs, inputs = parts[0], parts[1]
            outputs, inputs = outputs.strip(), inputs.strip()
            # Temporarily replace escaped spaces in names so that
            # the list of dependencies can be split.
            outputs, inputs = outputs.replace("\\ ", "\x00"), inputs.replace("\\ ", "\x00")

            for output in outputs.split():
                output = output.replace("\x00", " ")
                for input in inputs.split():
                    input = input.replace("\x00", " ")
                    if output not in deps:
                        deps[output] = []
                    deps[output].append(input)

        return deps

    def which(self, executable):
        """ Find executable in PATH.

        Args:
            executable (str): Name of executable to be found.

        Returns:
            str: Full path to the executable.
        """
        executable = self.expand(executable)
        path = self._env.get("PATH")

        if self._chroot:
            path = path.split(fs.pathsep) if path else []
            path = [os.path.join(self._chroot, p.lstrip(fs.sep)) for p in self._chroot_path] + path
            path = fs.pathsep.join(path)

        result = shutil.which(executable, path=path)
        if result and self._chroot and result.startswith(self._chroot):
            result = result[len(self._chroot):]
        return result

    def write_file(self, pathname, content=None, expand=True, **kwargs):
        """ Creates a file.

        Note:
            Existing files are overwritten.

        Args:
            pathname (str): Name/path of file to be created.
            content (str, optional): Data to be written to the file.
            expand (boolean, optional): Expand macros in file content.
               Default: True.
            **kwargs (dict, optional): Additional key value dictionary
                used in macro expansion.
        """
        pathname = self.expand_path(pathname)
        if content is None:
            content = ''
        if expand:
            content = self.expand(content, **kwargs)
        with open(pathname, "wb") as f:
            f.write(content.encode())

    @property
    def wsroot(self):
        """ Return the root path of all build directories """
        from jolt.loader import get_workspacedir
        return fs.path.normpath(get_workspacedir())



from contextlib import contextmanager
from datetime import datetime

from jolt import config
from jolt import filesystem as fs
from jolt import log
from jolt.error import raise_error_if
from jolt.hooks import TaskHook, TaskHookFactory


log.verbose("[LogStash] Loaded")


class LogStashHooks(TaskHook):
    def __init__(self):
        self._uri = config.get("logstash", "http.uri", "http://logstash")
        self._failed_enabled = config.getboolean("logstash", "failed", False)
        self._finished_enabled = config.getboolean("logstash", "passed", config.getboolean("logstash", "finished", False))
        raise_error_if(not self._uri, "logstash.http.uri not configured")

    def _get_uri(self, task):
        return "{}/{}-{}.txt".format(
            self._uri,
            datetime.now().strftime("%Y-%m-%d_%H%M%S.%f"),
            task.canonical_name)

    def _stash_log(self, task, logbuffer):
        with task.tools.tmpdir("logstash") as tmp:
            filepath = fs.path.join(tmp, "log")
            with open(filepath, "w") as f:
                f.write(logbuffer)
            task.logstash = self._get_uri(task)
            task.tools.upload(filepath, task.logstash, exceptions=False)

    @contextmanager
    def task_run(self, task):
        with log.threadsink() as logsink:
            try:
                yield
            except Exception as e:
                if self._failed_enabled:
                    self._stash_log(task, logsink.getvalue())
                raise e
            else:
                if self._finished_enabled:
                    self._stash_log(task, logsink.getvalue())


# Must run before other plugins which depend on the
# logstash_uri TaskProxy attribute.
@TaskHookFactory.register_with_prio(10)
class LogStashFactory(TaskHookFactory):
    def create(self, env):
        return LogStashHooks()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: jolt/plugins/remote_execution/worker.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from jolt import common_pb2 as jolt_dot_common__pb2
from jolt.plugins.remote_execution import scheduler_pb2 as jolt_dot_plugins_dot_remote__execution_dot_scheduler__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n*jolt/plugins/remote_execution/worker.proto\x1a\x11jolt/common.proto\x1a-jolt/plugins/remote_execution/scheduler.proto\"\xaf\x01\n\rWorkerRequest\x12%\n\x06\x61\x63tion\x18\x01 \x01(\x0e\x32\x15.WorkerRequest.Action\x12\x1c\n\x05\x62uild\x18\x02 \x01(\x0b\x32\r.BuildRequest\x12\x11\n\tworker_id\x18\x03 \x01(\t\x12\x10\n\x08\x62uild_id\x18\x04 \x01(\t\"4\n\x06\x41\x63tion\x12\t\n\x05\x42UILD\x10\x00\x12\x10\n\x0c\x43\x41NCEL_BUILD\x10\x01\x12\r\n\tTERMINATE\x10\x02\"\xf1\x01\n\x0cWorkerUpdate\x12$\n\x06status\x18\x01 \x01(\x0e\x32\x14.WorkerUpdate.Status\x12\x1b\n\x08platform\x18\x02 \x01(\x0b\x32\t.Platform\x12 \n\rtask_platform\x18\x04 \x01(\x0b\x32\t.Platform\x12\x1b\n\x05\x65rror\x18\x03 \x01(\x0b\x32\x0c.WorkerError\"_\n\x06Status\x12\r\n\tENLISTING\x10\x00\x12\r\n\tDELISTING\x10\x01\x12\x0f\n\x0b\x42UILD_ENDED\x10\x02\x12\x11\n\rDEPLOY_FAILED\x10\x03\x12\x13\n\x0f\x45XECUTOR_FAILED\x10\x04\"/\n\x0bWorkerError\x12\x0f\n\x07message\x18\x01 \x01(\t\x12\x0f\n\x07\x64\x65tails\x18\x02 \x01(\t2i\n\x06Worker\x12\x34\n\x0fGetInstructions\x12\r.WorkerUpdate\x1a\x0e.WorkerRequest(\x01\x30\x01\x12)\n\x08GetTasks\x12\x0b.TaskUpdate\x1a\x0c.TaskRequest(\x01\x30\x01\x42\x0eZ\x0cpkg/protocolb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'jolt.plugins.remote_execution.worker_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'Z\014pkg/protocol'
  _globals['_WORKERREQUEST']._serialized_start=113
  _globals['_WORKERREQUEST']._serialized_end=288
  _globals['_WORKERREQUEST_ACTION']._serialized_start=236
  _globals['_WORKERREQUEST_ACTION']._serialized_end=288
  _globals['_WORKERUPDATE']._serialized_start=291
  _globals['_WORKERUPDATE']._serialized_end=532
  _globals['_WORKERUPDATE_STATUS']._serialized_start=437
  _globals['_WORKERUPDATE_STATUS']._serialized_end=532
  _globals['_WORKERERROR']._serialized_start=534
  _globals['_WORKERERROR']._serialized_end=581
  _globals['_WORKER']._serialized_start=583
  _globals['_WORKER']._serialized_end=688
# @@protoc_insertion_point(module_scope)
